local mymodule = {}

-- if a table_creation_script does not create the named table or throw an exception then you will get an infinite loop, so be careful

-- List of each device that we manage
mymodule.provisioning_device_seq = {
	"CREATE SEQUENCE provisioning_device_seq",
}

-- Multi-to-multi mapping of devices to classes
mymodule.devices_to_classes = {
	"CREATE TABLE devices_to_classes (device_id INTEGER, class_id INTEGER REFERENCES provisioning_classes)",
	"CREATE INDEX d2c_device_idx ON devices_to_classes (device_id)",
	"CREATE INDEX d2c_class_idx ON devices_to_classes (class_id)",
	-- Need to enforce that devices do not contain more than one class in the same class group (using triggers)
	-- Theoretically should also check on updates and updates of classes, but not going to bother
	"CREATE OR REPLACE FUNCTION check_class_groups() RETURNS TRIGGER AS $$ \
		BEGIN \
			PERFORM * from provisioning_classes join devices_to_classes using(class_id) where device_id=NEW.device_id and class_group_id=(select class_group_id from provisioning_classes where class_id=NEW.class_id); \
			IF FOUND THEN \
				RAISE EXCEPTION 'Device cannot have multiple classes within the same class group'; \
			END IF; \
			RETURN NEW; \
		END; \
	$$ LANGUAGE plpgsql",
	"CREATE TRIGGER check_class_trigger BEFORE INSERT ON devices_to_classes FOR EACH ROW EXECUTE PROCEDURE check_class_groups()",
}

-- List of different class groups
mymodule.provisioning_class_groups = {
	"CREATE TABLE provisioning_class_groups (class_group_id SERIAL PRIMARY KEY, name VARCHAR(255) UNIQUE, label VARCHAR(255) UNIQUE, seq INTEGER)",
	"INSERT INTO provisioning_class_groups VALUES(default, 'device', 'Device Model', '1')",
	"INSERT INTO provisioning_class_groups VALUES(default, 'services', 'Services', '2')",
}

-- List of different device classes
mymodule.provisioning_classes = {
	"CREATE TABLE provisioning_classes (class_id SERIAL PRIMARY KEY, class_group_id INTEGER REFERENCES provisioning_class_groups, label VARCHAR(255), seq INTEGER)",
	"CREATE INDEX classes_group_idx ON provisioning_classes (class_group_id)",
	"CREATE INDEX classes_label_idx ON provisioning_classes (label)",
	-- Need to enforce that group/label combo is unique
	"CREATE UNIQUE INDEX classes_group_label_idx ON provisioning_classes (class_group_id, label)",
	"INSERT INTO provisioning_classes VALUES(default, (SELECT class_group_id FROM provisioning_class_groups WHERE name='device'), 'Polycom SoundPoint IP 321/331/335 SIP', '1')",
	"INSERT INTO provisioning_classes VALUES(default, (SELECT class_group_id FROM provisioning_class_groups WHERE name='device'), 'Polycom SoundPoint IP 450 SIP', '2')",
	"INSERT INTO provisioning_classes VALUES(default, (SELECT class_group_id FROM provisioning_class_groups WHERE name='device'), 'Polycom SoundPoint IP 550/560 SIP', '3')",
	"INSERT INTO provisioning_classes VALUES(default, (SELECT class_group_id FROM provisioning_class_groups WHERE name='device'), 'Polycom SoundPoint IP 650/670 SIP', '4')",
	"INSERT INTO provisioning_classes VALUES(default, (SELECT class_group_id FROM provisioning_class_groups WHERE name='device'), 'Polycom SoundStation IP 5000/6000/7000 SIP', '5')",
	"INSERT INTO provisioning_classes VALUES(default, (SELECT class_group_id FROM provisioning_class_groups WHERE name='device'), 'Polycom VVX 1500', '6')",
	"INSERT INTO provisioning_classes VALUES(default, (SELECT class_group_id FROM provisioning_class_groups WHERE name='device'), 'Polycom VVX 101', '7')",
	"INSERT INTO provisioning_classes VALUES(default, (SELECT class_group_id FROM provisioning_class_groups WHERE name='device'), 'Polycom VVX 201', '8')",
	"INSERT INTO provisioning_classes VALUES(default, (SELECT class_group_id FROM provisioning_class_groups WHERE name='device'), 'Polycom VVX 300/301/310/311', '9')",
	"INSERT INTO provisioning_classes VALUES(default, (SELECT class_group_id FROM provisioning_class_groups WHERE name='device'), 'Polycom VVX 400/401/410/411', '10')",
	"INSERT INTO provisioning_classes VALUES(default, (SELECT class_group_id FROM provisioning_class_groups WHERE name='device'), 'Polycom VVX 500/501', '11')",
	"INSERT INTO provisioning_classes VALUES(default, (SELECT class_group_id FROM provisioning_class_groups WHERE name='device'), 'Polycom VVX 600/601', '12')",
	"INSERT INTO provisioning_classes VALUES(default, (SELECT class_group_id FROM provisioning_class_groups WHERE name='device'), 'Linksys Internet Phone Adapter PAP2T', '100')",
	"INSERT INTO provisioning_classes VALUES(default, (SELECT class_group_id FROM provisioning_class_groups WHERE name='device'), 'Cisco 2 Port Phone Adapter SPA112', '200')",
	"INSERT INTO provisioning_classes VALUES(default, (SELECT class_group_id FROM provisioning_class_groups WHERE name='device'), 'Snom 300', '300')",
	"INSERT INTO provisioning_classes VALUES(default, (SELECT class_group_id FROM provisioning_class_groups WHERE name='device'), 'Snom 320/360/370', '301')",
	-- Leaving space in seq for future Grandstream phone support
	"INSERT INTO provisioning_classes VALUES(default, (SELECT class_group_id FROM provisioning_class_groups WHERE name='device'), 'Grandstream HT-502', '450')",
	"INSERT INTO provisioning_classes VALUES(default, (SELECT class_group_id FROM provisioning_class_groups WHERE name='device'), 'Grandstream HT701', '452')",
	"INSERT INTO provisioning_classes VALUES(default, (SELECT class_group_id FROM provisioning_class_groups WHERE name='device'), 'Grandstream HT702', '453')",
	"INSERT INTO provisioning_classes VALUES(default, (SELECT class_group_id FROM provisioning_class_groups WHERE name='device'), 'Grandstream HT704', '454')",
	"INSERT INTO provisioning_classes VALUES(default, (SELECT class_group_id FROM provisioning_class_groups WHERE name='device'), 'CyberData Intercom', '500')",
	"INSERT INTO provisioning_classes VALUES(default, (SELECT class_group_id FROM provisioning_class_groups WHERE name='device'), 'CyberData Paging Zone Controller', '501')",
	"INSERT INTO provisioning_classes VALUES(default, (SELECT class_group_id FROM provisioning_class_groups WHERE name='device'), 'Algo 8180 SIP Audio Alerter', '600')",
	"INSERT INTO provisioning_classes VALUES(default, (SELECT class_group_id FROM provisioning_class_groups WHERE name='device'), 'Unsupported Phone', '999')",
	-- services
	"INSERT INTO provisioning_classes VALUES(default, (SELECT class_group_id FROM provisioning_class_groups WHERE name='services'), 'Office Phone', '1')",
	"INSERT INTO provisioning_classes VALUES(default, (SELECT class_group_id FROM provisioning_class_groups WHERE name='services'), 'Residential Phone', '11')",
	"INSERT INTO provisioning_classes VALUES(default, (SELECT class_group_id FROM provisioning_class_groups WHERE name='services'), 'Public Phone', '21')",
	"INSERT INTO provisioning_classes VALUES(default, (SELECT class_group_id FROM provisioning_class_groups WHERE name='services'), 'Hotline', '31')",
}

-- Multi-to-multi mapping of classes to parameter groups
mymodule.classes_to_param_groups = {
	"CREATE TABLE classes_to_param_groups (class_id INTEGER REFERENCES provisioning_classes, group_id INTEGER REFERENCES provisioning_groups)",
	"CREATE INDEX c2g_class_idx ON classes_to_param_groups (class_id)",
	"CREATE INDEX c2g_group_idx ON classes_to_param_groups (group_id)",
	-- Need to enforce that classes do not contain more than one parameter group with same name (using triggers)
	-- Theoretically should also check on updates and updates of groups, but not going to bother
	"CREATE OR REPLACE FUNCTION check_group_names() RETURNS TRIGGER AS $$ \
	        BEGIN \
			PERFORM * from provisioning_groups join classes_to_param_groups using(group_id) where class_id=NEW.class_id and name=(select name from provisioning_groups where group_id=NEW.group_id); \
			IF FOUND THEN \
				RAISE EXCEPTION 'Classes cannot have multiple parameter groups with the same name'; \
			END IF; \
			RETURN NEW; \
		END; \
	$$ LANGUAGE plpgsql",
	"CREATE TRIGGER check_group_trigger BEFORE INSERT ON classes_to_param_groups FOR EACH ROW EXECUTE PROCEDURE check_group_names()",
	-- Devices
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom SoundPoint IP 321/331/335 SIP'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 1 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom SoundPoint IP 321/331/335 SIP'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 2 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom SoundPoint IP 321/331/335 SIP'), (SELECT group_id FROM provisioning_groups WHERE label='Polycom Device'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom SoundPoint IP 450 SIP'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 1 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom SoundPoint IP 450 SIP'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 2 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom SoundPoint IP 450 SIP'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 3 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom SoundPoint IP 450 SIP'), (SELECT group_id FROM provisioning_groups WHERE label='Polycom Device'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom SoundPoint IP 550/560 SIP'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 1 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom SoundPoint IP 550/560 SIP'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 2 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom SoundPoint IP 550/560 SIP'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 3 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom SoundPoint IP 550/560 SIP'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 4 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom SoundPoint IP 550/560 SIP'), (SELECT group_id FROM provisioning_groups WHERE label='Polycom Device'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom SoundPoint IP 650/670 SIP'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 1 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom SoundPoint IP 650/670 SIP'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 2 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom SoundPoint IP 650/670 SIP'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 3 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom SoundPoint IP 650/670 SIP'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 4 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom SoundPoint IP 650/670 SIP'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 5 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom SoundPoint IP 650/670 SIP'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 6 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom SoundPoint IP 650/670 SIP'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 7 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom SoundPoint IP 650/670 SIP'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 8 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom SoundPoint IP 650/670 SIP'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 9 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom SoundPoint IP 650/670 SIP'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 10 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom SoundPoint IP 650/670 SIP'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 11 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom SoundPoint IP 650/670 SIP'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 12 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom SoundPoint IP 650/670 SIP'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 13 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom SoundPoint IP 650/670 SIP'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 14 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom SoundPoint IP 650/670 SIP'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 15 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom SoundPoint IP 650/670 SIP'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 16 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom SoundPoint IP 650/670 SIP'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 17 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom SoundPoint IP 650/670 SIP'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 18 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom SoundPoint IP 650/670 SIP'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 19 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom SoundPoint IP 650/670 SIP'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 20 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom SoundPoint IP 650/670 SIP'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 21 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom SoundPoint IP 650/670 SIP'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 22 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom SoundPoint IP 650/670 SIP'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 23 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom SoundPoint IP 650/670 SIP'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 24 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom SoundPoint IP 650/670 SIP'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 25 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom SoundPoint IP 650/670 SIP'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 26 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom SoundPoint IP 650/670 SIP'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 27 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom SoundPoint IP 650/670 SIP'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 28 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom SoundPoint IP 650/670 SIP'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 29 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom SoundPoint IP 650/670 SIP'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 30 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom SoundPoint IP 650/670 SIP'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 31 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom SoundPoint IP 650/670 SIP'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 32 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom SoundPoint IP 650/670 SIP'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 33 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom SoundPoint IP 650/670 SIP'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 34 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom SoundPoint IP 650/670 SIP'), (SELECT group_id FROM provisioning_groups WHERE label='Polycom Device'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom SoundStation IP 5000/6000/7000 SIP'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 1 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom SoundStation IP 5000/6000/7000 SIP'), (SELECT group_id FROM provisioning_groups WHERE label='Polycom Device'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 101'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 1 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 101'), (SELECT group_id FROM provisioning_groups WHERE label='Polycom Device'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 201'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 1 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 201'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 2 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 201'), (SELECT group_id FROM provisioning_groups WHERE label='Polycom Device'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 300/301/310/311'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 1 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 300/301/310/311'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 2 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 300/301/310/311'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 3 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 300/301/310/311'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 4 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 300/301/310/311'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 5 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 300/301/310/311'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 6 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 300/301/310/311'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 7 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 300/301/310/311'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 8 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 300/301/310/311'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 9 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 300/301/310/311'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 10 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 300/301/310/311'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 11 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 300/301/310/311'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 12 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 300/301/310/311'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 13 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 300/301/310/311'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 14 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 300/301/310/311'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 15 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 300/301/310/311'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 16 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 300/301/310/311'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 17 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 300/301/310/311'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 18 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 300/301/310/311'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 19 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 300/301/310/311'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 20 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 300/301/310/311'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 21 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 300/301/310/311'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 22 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 300/301/310/311'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 23 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 300/301/310/311'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 24 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 300/301/310/311'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 25 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 300/301/310/311'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 26 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 300/301/310/311'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 27 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 300/301/310/311'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 28 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 300/301/310/311'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 29 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 300/301/310/311'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 30 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 300/301/310/311'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 31 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 300/301/310/311'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 32 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 300/301/310/311'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 33 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 300/301/310/311'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 34 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 300/301/310/311'), (SELECT group_id FROM provisioning_groups WHERE label='Polycom Device'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 400/401/410/411'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 1 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 400/401/410/411'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 2 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 400/401/410/411'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 3 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 400/401/410/411'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 4 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 400/401/410/411'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 5 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 400/401/410/411'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 6 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 400/401/410/411'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 7 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 400/401/410/411'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 8 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 400/401/410/411'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 9 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 400/401/410/411'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 10 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 400/401/410/411'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 11 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 400/401/410/411'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 12 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 400/401/410/411'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 13 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 400/401/410/411'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 14 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 400/401/410/411'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 15 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 400/401/410/411'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 16 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 400/401/410/411'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 17 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 400/401/410/411'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 18 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 400/401/410/411'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 19 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 400/401/410/411'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 20 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 400/401/410/411'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 21 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 400/401/410/411'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 22 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 400/401/410/411'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 23 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 400/401/410/411'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 24 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 400/401/410/411'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 25 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 400/401/410/411'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 26 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 400/401/410/411'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 27 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 400/401/410/411'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 28 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 400/401/410/411'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 29 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 400/401/410/411'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 30 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 400/401/410/411'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 31 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 400/401/410/411'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 32 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 400/401/410/411'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 33 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 400/401/410/411'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 34 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 400/401/410/411'), (SELECT group_id FROM provisioning_groups WHERE label='Polycom Device'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 500/501'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 1 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 500/501'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 2 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 500/501'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 3 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 500/501'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 4 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 500/501'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 5 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 500/501'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 6 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 500/501'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 7 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 500/501'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 8 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 500/501'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 9 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 500/501'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 10 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 500/501'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 11 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 500/501'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 12 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 500/501'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 13 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 500/501'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 14 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 500/501'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 15 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 500/501'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 16 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 500/501'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 17 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 500/501'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 18 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 500/501'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 19 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 500/501'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 20 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 500/501'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 21 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 500/501'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 22 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 500/501'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 23 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 500/501'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 24 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 500/501'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 25 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 500/501'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 26 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 500/501'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 27 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 500/501'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 28 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 500/501'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 29 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 500/501'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 30 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 500/501'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 31 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 500/501'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 32 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 500/501'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 33 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 500/501'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 34 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 500/501'), (SELECT group_id FROM provisioning_groups WHERE label='Polycom Device'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 600/601'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 1 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 600/601'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 2 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 600/601'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 3 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 600/601'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 4 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 600/601'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 5 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 600/601'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 6 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 600/601'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 7 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 600/601'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 8 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 600/601'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 9 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 600/601'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 10 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 600/601'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 11 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 600/601'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 12 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 600/601'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 13 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 600/601'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 14 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 600/601'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 15 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 600/601'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 16 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 600/601'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 17 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 600/601'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 18 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 600/601'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 19 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 600/601'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 20 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 600/601'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 21 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 600/601'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 22 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 600/601'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 23 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 600/601'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 24 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 600/601'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 25 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 600/601'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 26 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 600/601'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 27 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 600/601'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 28 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 600/601'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 29 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 600/601'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 30 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 600/601'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 31 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 600/601'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 32 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 600/601'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 33 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 600/601'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 34 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 600/601'), (SELECT group_id FROM provisioning_groups WHERE label='Polycom Device'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 1500'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 1 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 1500'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 2 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 1500'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 3 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 1500'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 4 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 1500'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 5 (Polycom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Polycom VVX 1500'), (SELECT group_id FROM provisioning_groups WHERE label='Polycom Device'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Linksys Internet Phone Adapter PAP2T'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 1 (Linksys)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Linksys Internet Phone Adapter PAP2T'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 2 (Linksys)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Linksys Internet Phone Adapter PAP2T'), (SELECT group_id FROM provisioning_groups WHERE label='Linksys ATA Device'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Cisco 2 Port Phone Adapter SPA112'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 1 (Cisco)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Cisco 2 Port Phone Adapter SPA112'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 2 (Cisco)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Cisco 2 Port Phone Adapter SPA112'), (SELECT group_id FROM provisioning_groups WHERE label='Cisco ATA Device'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Snom 300'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 1 (Snom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Snom 300'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 2 (Snom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Snom 300'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 3 (Snom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Snom 300'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 4 (Snom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Snom 300'), (SELECT group_id FROM provisioning_groups WHERE label='Snom Device'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Snom 320/360/370'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 1 (Snom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Snom 320/360/370'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 2 (Snom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Snom 320/360/370'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 3 (Snom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Snom 320/360/370'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 4 (Snom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Snom 320/360/370'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 5 (Snom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Snom 320/360/370'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 6 (Snom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Snom 320/360/370'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 7 (Snom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Snom 320/360/370'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 8 (Snom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Snom 320/360/370'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 9 (Snom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Snom 320/360/370'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 10 (Snom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Snom 320/360/370'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 11 (Snom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Snom 320/360/370'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 12 (Snom)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Snom 320/360/370'), (SELECT group_id FROM provisioning_groups WHERE label='Snom Device'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Algo 8180 SIP Audio Alerter'), (SELECT group_id FROM provisioning_groups WHERE label='Ring Extension 1 (Algo)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Algo 8180 SIP Audio Alerter'), (SELECT group_id FROM provisioning_groups WHERE label='Ring Extension 2 (Algo)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Algo 8180 SIP Audio Alerter'), (SELECT group_id FROM provisioning_groups WHERE label='Ring Extension 3 (Algo)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Algo 8180 SIP Audio Alerter'), (SELECT group_id FROM provisioning_groups WHERE label='Ring Extension 4 (Algo)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Algo 8180 SIP Audio Alerter'), (SELECT group_id FROM provisioning_groups WHERE label='Ring Extension 5 (Algo)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Algo 8180 SIP Audio Alerter'), (SELECT group_id FROM provisioning_groups WHERE label='Page Extension (Algo)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Algo 8180 SIP Audio Alerter'), (SELECT group_id FROM provisioning_groups WHERE label='Algo 8180 Device'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='CyberData Intercom'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 1 (CyberData)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='CyberData Intercom'), (SELECT group_id FROM provisioning_groups WHERE label='CyberData Intercom Device'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Grandstream HT-502'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 1 (Grandstream)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Grandstream HT-502'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 2 (Grandstream)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Grandstream HT-502'), (SELECT group_id FROM provisioning_groups WHERE label='Grandstream Device'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Grandstream HT701'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 1 (Grandstream)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Grandstream HT701'), (SELECT group_id FROM provisioning_groups WHERE label='Grandstream Device'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Grandstream HT702'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 1 (Grandstream)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Grandstream HT702'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 2 (Grandstream)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Grandstream HT702'), (SELECT group_id FROM provisioning_groups WHERE label='Grandstream Device'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Grandstream HT704'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 1 (Grandstream)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Grandstream HT704'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 2 (Grandstream)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Grandstream HT704'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 3 (Grandstream)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Grandstream HT704'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 4 (Grandstream)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Grandstream HT704'), (SELECT group_id FROM provisioning_groups WHERE label='Grandstream Device'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Unsupported Phone'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 1 (Unsupported)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='CyberData Paging Zone Controller'), (SELECT group_id FROM provisioning_groups WHERE label='Registration 1 (CyberData Paging Zone Controller)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='CyberData Paging Zone Controller'), (SELECT group_id FROM provisioning_groups WHERE label='CyberData Paging Zone Controller Device'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='CyberData Paging Zone Controller'), (SELECT group_id FROM provisioning_groups WHERE label='Zone 00 (CyberData Paging Zone Controller)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='CyberData Paging Zone Controller'), (SELECT group_id FROM provisioning_groups WHERE label='Zone 01 (CyberData Paging Zone Controller)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='CyberData Paging Zone Controller'), (SELECT group_id FROM provisioning_groups WHERE label='Zone 02 (CyberData Paging Zone Controller)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='CyberData Paging Zone Controller'), (SELECT group_id FROM provisioning_groups WHERE label='Zone 03 (CyberData Paging Zone Controller)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='CyberData Paging Zone Controller'), (SELECT group_id FROM provisioning_groups WHERE label='Zone 04 (CyberData Paging Zone Controller)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='CyberData Paging Zone Controller'), (SELECT group_id FROM provisioning_groups WHERE label='Zone 05 (CyberData Paging Zone Controller)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='CyberData Paging Zone Controller'), (SELECT group_id FROM provisioning_groups WHERE label='Zone 06 (CyberData Paging Zone Controller)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='CyberData Paging Zone Controller'), (SELECT group_id FROM provisioning_groups WHERE label='Zone 07 (CyberData Paging Zone Controller)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='CyberData Paging Zone Controller'), (SELECT group_id FROM provisioning_groups WHERE label='Zone 08 (CyberData Paging Zone Controller)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='CyberData Paging Zone Controller'), (SELECT group_id FROM provisioning_groups WHERE label='Zone 09 (CyberData Paging Zone Controller)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='CyberData Paging Zone Controller'), (SELECT group_id FROM provisioning_groups WHERE label='Zone 10 (CyberData Paging Zone Controller)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='CyberData Paging Zone Controller'), (SELECT group_id FROM provisioning_groups WHERE label='Zone 11 (CyberData Paging Zone Controller)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='CyberData Paging Zone Controller'), (SELECT group_id FROM provisioning_groups WHERE label='Zone 12 (CyberData Paging Zone Controller)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='CyberData Paging Zone Controller'), (SELECT group_id FROM provisioning_groups WHERE label='Zone 13 (CyberData Paging Zone Controller)'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='CyberData Paging Zone Controller'), (SELECT group_id FROM provisioning_groups WHERE label='Zone 14 (CyberData Paging Zone Controller)'))",
	-- Services
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Office Phone'), (SELECT group_id FROM provisioning_groups WHERE label='Office Phone'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Residential Phone'), (SELECT group_id FROM provisioning_groups WHERE label='Residential Phone'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Public Phone'), (SELECT group_id FROM provisioning_groups WHERE label='Public Phone'))",
	"INSERT INTO classes_to_param_groups VALUES((SELECT class_id FROM provisioning_classes WHERE label='Hotline'), (SELECT group_id FROM provisioning_groups WHERE label='Hotline'))",
}

-- List of different parameter groups
mymodule.provisioning_groups = {
	"CREATE TABLE provisioning_groups (group_id SERIAL PRIMARY KEY, name VARCHAR(255), label VARCHAR(255) UNIQUE, seq INTEGER)",
	"CREATE INDEX groups_name_idx ON provisioning_groups (name)",
	"INSERT INTO provisioning_groups VALUES(default, 'device', 'Linksys ATA Device', '1')",
	"INSERT INTO provisioning_groups VALUES(default, 'device', 'Cisco ATA Device', '1')",
	"INSERT INTO provisioning_groups VALUES(default, 'device', 'Polycom Device', '1')",
	"INSERT INTO provisioning_groups VALUES(default, 'device', 'Snom Device', '1')",
	"INSERT INTO provisioning_groups VALUES(default, 'device', 'Algo 8180 Device', '1')",
	"INSERT INTO provisioning_groups VALUES(default, 'device', 'CyberData Intercom Device', '1')",
	"INSERT INTO provisioning_groups VALUES(default, 'device', 'Grandstream Device', '1')",
	"INSERT INTO provisioning_groups VALUES(default, 'device', 'CyberData Paging Zone Controller Device', '1')",
	"INSERT INTO provisioning_groups VALUES(default, 'reg1', 'Registration 1 (Polycom)', '10')",
	"INSERT INTO provisioning_groups VALUES(default, 'reg1', 'Registration 1 (Linksys)', '10')",
	"INSERT INTO provisioning_groups VALUES(default, 'reg1', 'Registration 1 (Cisco)', '10')",
	"INSERT INTO provisioning_groups VALUES(default, 'reg1', 'Registration 1 (Unsupported)', '10')",
	"INSERT INTO provisioning_groups VALUES(default, 'reg1', 'Registration 1 (Snom)', '10')",
	"INSERT INTO provisioning_groups VALUES(default, 'reg1', 'Ring Extension 1 (Algo)', '10')",
	"INSERT INTO provisioning_groups VALUES(default, 'reg1', 'Registration 1 (CyberData)', '10')",
	"INSERT INTO provisioning_groups VALUES(default, 'reg1', 'Registration 1 (CyberData Paging Zone Controller)', '10')",
	"INSERT INTO provisioning_groups VALUES(default, 'reg1', 'Registration 1 (Grandstream)', '10')",
	"INSERT INTO provisioning_groups VALUES(default, 'reg2', 'Registration 2 (Polycom)', '10')",
	"INSERT INTO provisioning_groups VALUES(default, 'reg2', 'Registration 2 (Linksys)', '10')",
	"INSERT INTO provisioning_groups VALUES(default, 'reg2', 'Registration 2 (Cisco)', '10')",
	"INSERT INTO provisioning_groups VALUES(default, 'reg2', 'Registration 2 (Snom)', '10')",
	"INSERT INTO provisioning_groups VALUES(default, 'reg2', 'Ring Extension 2 (Algo)', '10')",
	"INSERT INTO provisioning_groups VALUES(default, 'reg2', 'Registration 2 (Grandstream)', '10')",
	"INSERT INTO provisioning_groups VALUES(default, 'reg3', 'Registration 3 (Polycom)', '10')",
	"INSERT INTO provisioning_groups VALUES(default, 'reg3', 'Registration 3 (Snom)', '10')",
	"INSERT INTO provisioning_groups VALUES(default, 'reg3', 'Ring Extension 3 (Algo)', '10')",
	"INSERT INTO provisioning_groups VALUES(default, 'reg3', 'Registration 3 (Grandstream)', '10')",
	"INSERT INTO provisioning_groups VALUES(default, 'reg4', 'Registration 4 (Polycom)', '10')",
	"INSERT INTO provisioning_groups VALUES(default, 'reg4', 'Registration 4 (Snom)', '10')",
	"INSERT INTO provisioning_groups VALUES(default, 'reg4', 'Ring Extension 4 (Algo)', '10')",
	"INSERT INTO provisioning_groups VALUES(default, 'reg4', 'Registration 4 (Grandstream)', '10')",
	"INSERT INTO provisioning_groups VALUES(default, 'reg5', 'Registration 5 (Polycom)', '10')",
	"INSERT INTO provisioning_groups VALUES(default, 'reg5', 'Registration 5 (Snom)', '10')",
	"INSERT INTO provisioning_groups VALUES(default, 'reg5', 'Ring Extension 5 (Algo)', '10')",
	"INSERT INTO provisioning_groups VALUES(default, 'reg6', 'Registration 6 (Polycom)', '10')",
	"INSERT INTO provisioning_groups VALUES(default, 'reg6', 'Registration 6 (Snom)', '10')",
	"INSERT INTO provisioning_groups VALUES(default, 'reg6', 'Page Extension (Algo)', '10')",
	"INSERT INTO provisioning_groups VALUES(default, 'reg7', 'Registration 7 (Polycom)', '10')",
	"INSERT INTO provisioning_groups VALUES(default, 'reg7', 'Registration 7 (Snom)', '10')",
	"INSERT INTO provisioning_groups VALUES(default, 'reg8', 'Registration 8 (Polycom)', '10')",
	"INSERT INTO provisioning_groups VALUES(default, 'reg8', 'Registration 8 (Snom)', '10')",
	"INSERT INTO provisioning_groups VALUES(default, 'reg9', 'Registration 9 (Polycom)', '10')",
	"INSERT INTO provisioning_groups VALUES(default, 'reg9', 'Registration 9 (Snom)', '10')",
	"INSERT INTO provisioning_groups VALUES(default, 'reg10', 'Registration 10 (Polycom)', '11')",
	"INSERT INTO provisioning_groups VALUES(default, 'reg10', 'Registration 10 (Snom)', '11')",
	"INSERT INTO provisioning_groups VALUES(default, 'reg11', 'Registration 11 (Polycom)', '11')",
	"INSERT INTO provisioning_groups VALUES(default, 'reg11', 'Registration 11 (Snom)', '11')",
	"INSERT INTO provisioning_groups VALUES(default, 'reg12', 'Registration 12 (Polycom)', '11')",
	"INSERT INTO provisioning_groups VALUES(default, 'reg12', 'Registration 12 (Snom)', '11')",
	"INSERT INTO provisioning_groups VALUES(default, 'reg13', 'Registration 13 (Polycom)', '11')",
	"INSERT INTO provisioning_groups VALUES(default, 'reg14', 'Registration 14 (Polycom)', '11')",
	"INSERT INTO provisioning_groups VALUES(default, 'reg15', 'Registration 15 (Polycom)', '11')",
	"INSERT INTO provisioning_groups VALUES(default, 'reg16', 'Registration 16 (Polycom)', '11')",
	"INSERT INTO provisioning_groups VALUES(default, 'reg17', 'Registration 17 (Polycom)', '11')",
	"INSERT INTO provisioning_groups VALUES(default, 'reg18', 'Registration 18 (Polycom)', '11')",
	"INSERT INTO provisioning_groups VALUES(default, 'reg19', 'Registration 19 (Polycom)', '11')",
	"INSERT INTO provisioning_groups VALUES(default, 'reg20', 'Registration 20 (Polycom)', '12')",
	"INSERT INTO provisioning_groups VALUES(default, 'reg21', 'Registration 21 (Polycom)', '12')",
	"INSERT INTO provisioning_groups VALUES(default, 'reg22', 'Registration 22 (Polycom)', '12')",
	"INSERT INTO provisioning_groups VALUES(default, 'reg23', 'Registration 23 (Polycom)', '12')",
	"INSERT INTO provisioning_groups VALUES(default, 'reg24', 'Registration 24 (Polycom)', '12')",
	"INSERT INTO provisioning_groups VALUES(default, 'reg25', 'Registration 25 (Polycom)', '12')",
	"INSERT INTO provisioning_groups VALUES(default, 'reg26', 'Registration 26 (Polycom)', '12')",
	"INSERT INTO provisioning_groups VALUES(default, 'reg27', 'Registration 27 (Polycom)', '12')",
	"INSERT INTO provisioning_groups VALUES(default, 'reg28', 'Registration 28 (Polycom)', '12')",
	"INSERT INTO provisioning_groups VALUES(default, 'reg29', 'Registration 29 (Polycom)', '12')",
	"INSERT INTO provisioning_groups VALUES(default, 'reg30', 'Registration 30 (Polycom)', '13')",
	"INSERT INTO provisioning_groups VALUES(default, 'reg31', 'Registration 31 (Polycom)', '13')",
	"INSERT INTO provisioning_groups VALUES(default, 'reg32', 'Registration 32 (Polycom)', '13')",
	"INSERT INTO provisioning_groups VALUES(default, 'reg33', 'Registration 33 (Polycom)', '13')",
	"INSERT INTO provisioning_groups VALUES(default, 'reg34', 'Registration 34 (Polycom)', '13')",
	"INSERT INTO provisioning_groups VALUES(default, 'zone00', 'Zone 00 (CyberData Paging Zone Controller)', '14')",
	"INSERT INTO provisioning_groups VALUES(default, 'zone01', 'Zone 01 (CyberData Paging Zone Controller)', '14')",
	"INSERT INTO provisioning_groups VALUES(default, 'zone02', 'Zone 02 (CyberData Paging Zone Controller)', '14')",
	"INSERT INTO provisioning_groups VALUES(default, 'zone03', 'Zone 03 (CyberData Paging Zone Controller)', '14')",
	"INSERT INTO provisioning_groups VALUES(default, 'zone04', 'Zone 04 (CyberData Paging Zone Controller)', '14')",
	"INSERT INTO provisioning_groups VALUES(default, 'zone05', 'Zone 05 (CyberData Paging Zone Controller)', '14')",
	"INSERT INTO provisioning_groups VALUES(default, 'zone06', 'Zone 06 (CyberData Paging Zone Controller)', '14')",
	"INSERT INTO provisioning_groups VALUES(default, 'zone07', 'Zone 07 (CyberData Paging Zone Controller)', '14')",
	"INSERT INTO provisioning_groups VALUES(default, 'zone08', 'Zone 08 (CyberData Paging Zone Controller)', '14')",
	"INSERT INTO provisioning_groups VALUES(default, 'zone09', 'Zone 09 (CyberData Paging Zone Controller)', '14')",
	"INSERT INTO provisioning_groups VALUES(default, 'zone10', 'Zone 10 (CyberData Paging Zone Controller)', '14')",
	"INSERT INTO provisioning_groups VALUES(default, 'zone11', 'Zone 11 (CyberData Paging Zone Controller)', '14')",
	"INSERT INTO provisioning_groups VALUES(default, 'zone12', 'Zone 12 (CyberData Paging Zone Controller)', '14')",
	"INSERT INTO provisioning_groups VALUES(default, 'zone13', 'Zone 13 (CyberData Paging Zone Controller)', '14')",
	"INSERT INTO provisioning_groups VALUES(default, 'zone14', 'Zone 14 (CyberData Paging Zone Controller)', '14')",
	-- services
	"INSERT INTO provisioning_groups VALUES(default, 'services', 'Office Phone', '21')",
	"INSERT INTO provisioning_groups VALUES(default, 'services', 'Residential Phone', '31')",
	"INSERT INTO provisioning_groups VALUES(default, 'services', 'Public Phone', '41')",
	"INSERT INTO provisioning_groups VALUES(default, 'services', 'Hotline', '51')",
}

-- Multi-to-multi mapping of parameter groups to parameters
mymodule.param_groups_to_params = {
	"CREATE TABLE param_groups_to_params (group_id INTEGER REFERENCES provisioning_groups, param_id INTEGER REFERENCES provisioning_params, value VARCHAR(255), editable BOOLEAN)",
	"CREATE INDEX g2p_group_idx ON param_groups_to_params (group_id)",
	"CREATE INDEX g2p_param_idx ON param_groups_to_params (param_id)",
	"CREATE INDEX g2p_editable_idx ON param_groups_to_params (editable)",
	"CREATE UNIQUE INDEX g2p_group_param_idx ON param_groups_to_params (group_id, param_id)",
	-- Device
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Linksys ATA Device'), (SELECT param_id FROM provisioning_params WHERE name='template'), '/etc/provisioning/templates/linksysata-template.lua', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Linksys ATA Device'), (SELECT param_id FROM provisioning_params WHERE name='mac'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Linksys ATA Device'), (SELECT param_id FROM provisioning_params WHERE name='registrar'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Linksys ATA Device'), (SELECT param_id FROM provisioning_params WHERE name='digitmap'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Linksys ATA Device'), (SELECT param_id FROM provisioning_params WHERE name='sntpserver'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Linksys ATA Device'), (SELECT param_id FROM provisioning_params WHERE name='timezone'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Linksys ATA Device'), (SELECT param_id FROM provisioning_params WHERE name='musiconhold'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Linksys ATA Device'), (SELECT param_id FROM provisioning_params WHERE name='adminpassword'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Cisco ATA Device'), (SELECT param_id FROM provisioning_params WHERE name='template'), '/etc/provisioning/templates/ciscoata-template.lua', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Cisco ATA Device'), (SELECT param_id FROM provisioning_params WHERE name='mac'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Cisco ATA Device'), (SELECT param_id FROM provisioning_params WHERE name='registrar'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Cisco ATA Device'), (SELECT param_id FROM provisioning_params WHERE name='digitmap'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Cisco ATA Device'), (SELECT param_id FROM provisioning_params WHERE name='sntpserver'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Cisco ATA Device'), (SELECT param_id FROM provisioning_params WHERE name='timezone'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Cisco ATA Device'), (SELECT param_id FROM provisioning_params WHERE name='musiconhold'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Cisco ATA Device'), (SELECT param_id FROM provisioning_params WHERE name='adminpassword'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Polycom Device'), (SELECT param_id FROM provisioning_params WHERE name='template'), '/etc/provisioning/templates/polycom-template.lua', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Polycom Device'), (SELECT param_id FROM provisioning_params WHERE name='mac'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Polycom Device'), (SELECT param_id FROM provisioning_params WHERE name='registrar'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Polycom Device'), (SELECT param_id FROM provisioning_params WHERE name='digitmap'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Polycom Device'), (SELECT param_id FROM provisioning_params WHERE name='digitmaptimeout'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Polycom Device'), (SELECT param_id FROM provisioning_params WHERE name='sntpserver'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Polycom Device'), (SELECT param_id FROM provisioning_params WHERE name='timezone'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Polycom Device'), (SELECT param_id FROM provisioning_params WHERE name='musiconhold'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Polycom Device'), (SELECT param_id FROM provisioning_params WHERE name='homepage'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Polycom Device'), (SELECT param_id FROM provisioning_params WHERE name='adminpassword'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Polycom Device'), (SELECT param_id FROM provisioning_params WHERE name='pcportenable'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Polycom Device'), (SELECT param_id FROM provisioning_params WHERE name='urldialingenable'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Polycom Device'), (SELECT param_id FROM provisioning_params WHERE name='militarytimeenable'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Polycom Device'), (SELECT param_id FROM provisioning_params WHERE name='clockenable'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Polycom Device'), (SELECT param_id FROM provisioning_params WHERE name='displaylang'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Polycom Device'), (SELECT param_id FROM provisioning_params WHERE name='dectenable'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Snom Device'), (SELECT param_id FROM provisioning_params WHERE name='template'), '/etc/provisioning/templates/snom-template.lua', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Snom Device'), (SELECT param_id FROM provisioning_params WHERE name='mac'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Snom Device'), (SELECT param_id FROM provisioning_params WHERE name='registrar'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Snom Device'), (SELECT param_id FROM provisioning_params WHERE name='digitmap'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Snom Device'), (SELECT param_id FROM provisioning_params WHERE name='digitmaptimeout'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Snom Device'), (SELECT param_id FROM provisioning_params WHERE name='sntpserver'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Snom Device'), (SELECT param_id FROM provisioning_params WHERE name='timezone'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Snom Device'), (SELECT param_id FROM provisioning_params WHERE name='musiconhold'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Snom Device'), (SELECT param_id FROM provisioning_params WHERE name='adminpassword'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Snom Device'), (SELECT param_id FROM provisioning_params WHERE name='pcportenable'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Snom Device'), (SELECT param_id FROM provisioning_params WHERE name='urldialingenable'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Snom Device'), (SELECT param_id FROM provisioning_params WHERE name='forwardnoanswer'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Snom Device'), (SELECT param_id FROM provisioning_params WHERE name='forwardnoanswerenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Snom Device'), (SELECT param_id FROM provisioning_params WHERE name='forwardbusy'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Snom Device'), (SELECT param_id FROM provisioning_params WHERE name='forwardbusyenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Snom Device'), (SELECT param_id FROM provisioning_params WHERE name='forwardall'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Snom Device'), (SELECT param_id FROM provisioning_params WHERE name='forwardallenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Algo 8180 Device'), (SELECT param_id FROM provisioning_params WHERE name='template'), '/etc/provisioning/templates/algo8180-template.lua', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Algo 8180 Device'), (SELECT param_id FROM provisioning_params WHERE name='mac'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Algo 8180 Device'), (SELECT param_id FROM provisioning_params WHERE name='registrar'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Algo 8180 Device'), (SELECT param_id FROM provisioning_params WHERE name='sntpserver'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Algo 8180 Device'), (SELECT param_id FROM provisioning_params WHERE name='adminpassword'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Algo 8180 Device'), (SELECT param_id FROM provisioning_params WHERE name='algoringvol'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Algo 8180 Device'), (SELECT param_id FROM provisioning_params WHERE name='algopagevol'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Algo 8180 Device'), (SELECT param_id FROM provisioning_params WHERE name='algospkmode'), null, true)";
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='CyberData Intercom Device'), (SELECT param_id FROM provisioning_params WHERE name='template'), '/etc/provisioning/templates/cyberdata-template.lua', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='CyberData Intercom Device'), (SELECT param_id FROM provisioning_params WHERE name='mac'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='CyberData Intercom Device'), (SELECT param_id FROM provisioning_params WHERE name='registrar'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='CyberData Intercom Device'), (SELECT param_id FROM provisioning_params WHERE name='sntpserver'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='CyberData Intercom Device'), (SELECT param_id FROM provisioning_params WHERE name='timezone'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='CyberData Intercom Device'), (SELECT param_id FROM provisioning_params WHERE name='cyberspeakervolume'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='CyberData Intercom Device'), (SELECT param_id FROM provisioning_params WHERE name='adminpassword'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='CyberData Intercom Device'), (SELECT param_id FROM provisioning_params WHERE name='cyberringvolume'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='CyberData Intercom Device'), (SELECT param_id FROM provisioning_params WHERE name='cybermicgain'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='CyberData Intercom Device'), (SELECT param_id FROM provisioning_params WHERE name='activaterelaywithdtmf'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='CyberData Intercom Device'), (SELECT param_id FROM provisioning_params WHERE name='dtmfactivationcode'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='CyberData Intercom Device'), (SELECT param_id FROM provisioning_params WHERE name='dtmfactivationduration'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='CyberData Intercom Device'), (SELECT param_id FROM provisioning_params WHERE name='playtonewhilerelayactive'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='CyberData Intercom Device'), (SELECT param_id FROM provisioning_params WHERE name='activaterelayduringring'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='CyberData Intercom Device'), (SELECT param_id FROM provisioning_params WHERE name='activaterelayduringcall'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='CyberData Intercom Device'), (SELECT param_id FROM provisioning_params WHERE name='autoanswerincomingcalls'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='CyberData Intercom Device'), (SELECT param_id FROM provisioning_params WHERE name='sensornormallyclosed'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='CyberData Intercom Device'), (SELECT param_id FROM provisioning_params WHERE name='sensortimeout'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='CyberData Intercom Device'), (SELECT param_id FROM provisioning_params WHERE name='sensorflashled'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='CyberData Intercom Device'), (SELECT param_id FROM provisioning_params WHERE name='sensoractivaterelay'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='CyberData Intercom Device'), (SELECT param_id FROM provisioning_params WHERE name='sensorplaylocally'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='CyberData Intercom Device'), (SELECT param_id FROM provisioning_params WHERE name='sensorcall'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='CyberData Intercom Device'), (SELECT param_id FROM provisioning_params WHERE name='sensordialoutextension'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='CyberData Intercom Device'), (SELECT param_id FROM provisioning_params WHERE name='sensordialoutid'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='CyberData Intercom Device'), (SELECT param_id FROM provisioning_params WHERE name='sensorrepeat'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='CyberData Intercom Device'), (SELECT param_id FROM provisioning_params WHERE name='sensorplayremotely'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Grandstream Device'), (SELECT param_id FROM provisioning_params WHERE name='template'), '/etc/provisioning/templates/grandstream-template.lua', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Grandstream Device'), (SELECT param_id FROM provisioning_params WHERE name='mac'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Grandstream Device'), (SELECT param_id FROM provisioning_params WHERE name='registrar'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Grandstream Device'), (SELECT param_id FROM provisioning_params WHERE name='sntpserver'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Grandstream Device'), (SELECT param_id FROM provisioning_params WHERE name='timezone'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Grandstream Device'), (SELECT param_id FROM provisioning_params WHERE name='adminpassword'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='CyberData Paging Zone Controller Device'), (SELECT param_id FROM provisioning_params WHERE name='registrar'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='CyberData Paging Zone Controller Device'), (SELECT param_id FROM provisioning_params WHERE name='sntpserver'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='CyberData Paging Zone Controller Device'), (SELECT param_id FROM provisioning_params WHERE name='timezone'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='CyberData Paging Zone Controller Device'), (SELECT param_id FROM provisioning_params WHERE name='mac'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='CyberData Paging Zone Controller Device'), (SELECT param_id FROM provisioning_params WHERE name='adminpassword'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='CyberData Paging Zone Controller Device'), (SELECT param_id FROM provisioning_params WHERE name='template'), '/etc/provisioning/templates/cyberdata-template.lua', 'f')",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='CyberData Paging Zone Controller Device'), (SELECT param_id FROM provisioning_params WHERE name='bypassdtmf'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='CyberData Paging Zone Controller Device'), (SELECT param_id FROM provisioning_params WHERE name='beeponinitialization'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='CyberData Paging Zone Controller Device'), (SELECT param_id FROM provisioning_params WHERE name='beepbeforepage'), null, true)",
	-- Registrations
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 1 (Unsupported)'), (SELECT param_id FROM provisioning_params WHERE name='extension'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 1 (Unsupported)'), (SELECT param_id FROM provisioning_params WHERE name='password'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 1 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='extension'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 1 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='password'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 1 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardnoanswer'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 1 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardnoanswerenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 1 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardbusy'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 1 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardbusyenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 1 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardall'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 1 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardallenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 1 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='callerid'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 1 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='polycomringtone'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 1 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='callwaitingenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 1 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='scaenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 1 (Linksys)'), (SELECT param_id FROM provisioning_params WHERE name='extension'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 1 (Linksys)'), (SELECT param_id FROM provisioning_params WHERE name='password'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 1 (Linksys)'), (SELECT param_id FROM provisioning_params WHERE name='forwardnoanswer'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 1 (Linksys)'), (SELECT param_id FROM provisioning_params WHERE name='forwardnoanswerenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 1 (Linksys)'), (SELECT param_id FROM provisioning_params WHERE name='forwardbusy'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 1 (Linksys)'), (SELECT param_id FROM provisioning_params WHERE name='forwardbusyenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 1 (Linksys)'), (SELECT param_id FROM provisioning_params WHERE name='forwardall'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 1 (Linksys)'), (SELECT param_id FROM provisioning_params WHERE name='forwardallenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 1 (Linksys)'), (SELECT param_id FROM provisioning_params WHERE name='callerid'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 1 (Cisco)'), (SELECT param_id FROM provisioning_params WHERE name='extension'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 1 (Cisco)'), (SELECT param_id FROM provisioning_params WHERE name='password'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 1 (Cisco)'), (SELECT param_id FROM provisioning_params WHERE name='forwardnoanswer'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 1 (Cisco)'), (SELECT param_id FROM provisioning_params WHERE name='forwardnoanswerenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 1 (Cisco)'), (SELECT param_id FROM provisioning_params WHERE name='forwardbusy'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 1 (Cisco)'), (SELECT param_id FROM provisioning_params WHERE name='forwardbusyenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 1 (Cisco)'), (SELECT param_id FROM provisioning_params WHERE name='forwardall'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 1 (Cisco)'), (SELECT param_id FROM provisioning_params WHERE name='forwardallenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 1 (Cisco)'), (SELECT param_id FROM provisioning_params WHERE name='callerid'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 1 (Snom)'), (SELECT param_id FROM provisioning_params WHERE name='extension'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 1 (Snom)'), (SELECT param_id FROM provisioning_params WHERE name='password'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 1 (Snom)'), (SELECT param_id FROM provisioning_params WHERE name='callerid'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Ring Extension 1 (Algo)'), (SELECT param_id FROM provisioning_params WHERE name='extension'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Ring Extension 1 (Algo)'), (SELECT param_id FROM provisioning_params WHERE name='password'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Ring Extension 1 (Algo)'), (SELECT param_id FROM provisioning_params WHERE name='algoringtone'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 1 (CyberData)'), (SELECT param_id FROM provisioning_params WHERE name='extension'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 1 (CyberData)'), (SELECT param_id FROM provisioning_params WHERE name='password'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 1 (CyberData)'), (SELECT param_id FROM provisioning_params WHERE name='callerid'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 1 (CyberData)'), (SELECT param_id FROM provisioning_params WHERE name='hotlineenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 1 (CyberData)'), (SELECT param_id FROM provisioning_params WHERE name='hotlinedestination'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 1 (CyberData Paging Zone Controller)'), (SELECT param_id FROM provisioning_params WHERE name='callerid'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 1 (CyberData Paging Zone Controller)'), (SELECT param_id FROM provisioning_params WHERE name='extension'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 1 (CyberData Paging Zone Controller)'), (SELECT param_id FROM provisioning_params WHERE name='password'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 1 (Grandstream)'), (SELECT param_id FROM provisioning_params WHERE name='extension'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 1 (Grandstream)'), (SELECT param_id FROM provisioning_params WHERE name='password'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 1 (Grandstream)'), (SELECT param_id FROM provisioning_params WHERE name='callerid'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 2 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='extension'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 2 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='password'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 2 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardnoanswer'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 2 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardnoanswerenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 2 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardbusy'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 2 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardbusyenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 2 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardall'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 2 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardallenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 2 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='callerid'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 2 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='polycomringtone'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 2 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='callwaitingenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 2 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='scaenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 2 (Linksys)'), (SELECT param_id FROM provisioning_params WHERE name='extension'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 2 (Linksys)'), (SELECT param_id FROM provisioning_params WHERE name='password'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 2 (Linksys)'), (SELECT param_id FROM provisioning_params WHERE name='forwardnoanswer'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 2 (Linksys)'), (SELECT param_id FROM provisioning_params WHERE name='forwardnoanswerenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 2 (Linksys)'), (SELECT param_id FROM provisioning_params WHERE name='forwardbusy'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 2 (Linksys)'), (SELECT param_id FROM provisioning_params WHERE name='forwardbusyenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 2 (Linksys)'), (SELECT param_id FROM provisioning_params WHERE name='forwardall'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 2 (Linksys)'), (SELECT param_id FROM provisioning_params WHERE name='forwardallenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 2 (Linksys)'), (SELECT param_id FROM provisioning_params WHERE name='callerid'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 2 (Cisco)'), (SELECT param_id FROM provisioning_params WHERE name='extension'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 2 (Cisco)'), (SELECT param_id FROM provisioning_params WHERE name='password'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 2 (Cisco)'), (SELECT param_id FROM provisioning_params WHERE name='forwardnoanswer'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 2 (Cisco)'), (SELECT param_id FROM provisioning_params WHERE name='forwardnoanswerenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 2 (Cisco)'), (SELECT param_id FROM provisioning_params WHERE name='forwardbusy'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 2 (Cisco)'), (SELECT param_id FROM provisioning_params WHERE name='forwardbusyenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 2 (Cisco)'), (SELECT param_id FROM provisioning_params WHERE name='forwardall'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 2 (Cisco)'), (SELECT param_id FROM provisioning_params WHERE name='forwardallenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 2 (Cisco)'), (SELECT param_id FROM provisioning_params WHERE name='callerid'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 2 (Snom)'), (SELECT param_id FROM provisioning_params WHERE name='extension'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 2 (Snom)'), (SELECT param_id FROM provisioning_params WHERE name='password'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 2 (Snom)'), (SELECT param_id FROM provisioning_params WHERE name='callerid'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Ring Extension 2 (Algo)'), (SELECT param_id FROM provisioning_params WHERE name='extension'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Ring Extension 2 (Algo)'), (SELECT param_id FROM provisioning_params WHERE name='password'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Ring Extension 2 (Algo)'), (SELECT param_id FROM provisioning_params WHERE name='algoringtone'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 2 (Grandstream)'), (SELECT param_id FROM provisioning_params WHERE name='extension'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 2 (Grandstream)'), (SELECT param_id FROM provisioning_params WHERE name='password'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 2 (Grandstream)'), (SELECT param_id FROM provisioning_params WHERE name='callerid'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 3 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='extension'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 3 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='password'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 3 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardnoanswer'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 3 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardnoanswerenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 3 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardbusy'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 3 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardbusyenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 3 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardall'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 3 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardallenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 3 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='callerid'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 3 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='polycomringtone'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 3 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='callwaitingenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 3 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='scaenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 3 (Snom)'), (SELECT param_id FROM provisioning_params WHERE name='extension'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 3 (Snom)'), (SELECT param_id FROM provisioning_params WHERE name='password'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 3 (Snom)'), (SELECT param_id FROM provisioning_params WHERE name='callerid'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Ring Extension 3 (Algo)'), (SELECT param_id FROM provisioning_params WHERE name='extension'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Ring Extension 3 (Algo)'), (SELECT param_id FROM provisioning_params WHERE name='password'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Ring Extension 3 (Algo)'), (SELECT param_id FROM provisioning_params WHERE name='algoringtone'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 4 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='extension'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 3 (Grandstream)'), (SELECT param_id FROM provisioning_params WHERE name='extension'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 3 (Grandstream)'), (SELECT param_id FROM provisioning_params WHERE name='password'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 3 (Grandstream)'), (SELECT param_id FROM provisioning_params WHERE name='callerid'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 4 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='password'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 4 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardnoanswer'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 4 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardnoanswerenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 4 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardbusy'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 4 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardbusyenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 4 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardall'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 4 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardallenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 4 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='callerid'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 4 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='polycomringtone'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 4 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='callwaitingenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 4 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='scaenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 4 (Snom)'), (SELECT param_id FROM provisioning_params WHERE name='extension'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 4 (Snom)'), (SELECT param_id FROM provisioning_params WHERE name='password'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 4 (Snom)'), (SELECT param_id FROM provisioning_params WHERE name='callerid'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Ring Extension 4 (Algo)'), (SELECT param_id FROM provisioning_params WHERE name='extension'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Ring Extension 4 (Algo)'), (SELECT param_id FROM provisioning_params WHERE name='password'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Ring Extension 4 (Algo)'), (SELECT param_id FROM provisioning_params WHERE name='algoringtone'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 4 (Grandstream)'), (SELECT param_id FROM provisioning_params WHERE name='extension'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 4 (Grandstream)'), (SELECT param_id FROM provisioning_params WHERE name='password'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 4 (Grandstream)'), (SELECT param_id FROM provisioning_params WHERE name='callerid'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 5 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='extension'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 5 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='password'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 5 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardnoanswer'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 5 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardnoanswerenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 5 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardbusy'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 5 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardbusyenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 5 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardall'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 5 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardallenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 5 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='callerid'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 5 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='polycomringtone'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 5 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='callwaitingenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 5 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='scaenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 5 (Snom)'), (SELECT param_id FROM provisioning_params WHERE name='extension'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 5 (Snom)'), (SELECT param_id FROM provisioning_params WHERE name='password'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 5 (Snom)'), (SELECT param_id FROM provisioning_params WHERE name='callerid'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Ring Extension 5 (Algo)'), (SELECT param_id FROM provisioning_params WHERE name='extension'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Ring Extension 5 (Algo)'), (SELECT param_id FROM provisioning_params WHERE name='password'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Ring Extension 5 (Algo)'), (SELECT param_id FROM provisioning_params WHERE name='algoringtone'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 6 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='extension'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 6 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='password'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 6 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardnoanswer'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 6 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardnoanswerenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 6 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardbusy'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 6 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardbusyenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 6 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardall'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 6 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardallenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 6 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='callerid'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 6 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='polycomringtone'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 6 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='callwaitingenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 6 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='scaenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 6 (Snom)'), (SELECT param_id FROM provisioning_params WHERE name='extension'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 6 (Snom)'), (SELECT param_id FROM provisioning_params WHERE name='password'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 6 (Snom)'), (SELECT param_id FROM provisioning_params WHERE name='callerid'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Page Extension (Algo)'), (SELECT param_id FROM provisioning_params WHERE name='extension'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Page Extension (Algo)'), (SELECT param_id FROM provisioning_params WHERE name='password'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Page Extension (Algo)'), (SELECT param_id FROM provisioning_params WHERE name='algopagetone'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 7 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='extension'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 7 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='password'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 7 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardnoanswer'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 7 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardnoanswerenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 7 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardbusy'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 7 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardbusyenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 7 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardall'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 7 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardallenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 7 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='callerid'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 7 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='polycomringtone'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 7 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='callwaitingenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 7 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='scaenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 7 (Snom)'), (SELECT param_id FROM provisioning_params WHERE name='extension'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 7 (Snom)'), (SELECT param_id FROM provisioning_params WHERE name='password'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 7 (Snom)'), (SELECT param_id FROM provisioning_params WHERE name='callerid'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 8 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='extension'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 8 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='password'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 8 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardnoanswer'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 8 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardnoanswerenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 8 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardbusy'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 8 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardbusyenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 8 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardall'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 8 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardallenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 8 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='callerid'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 8 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='polycomringtone'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 8 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='callwaitingenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 8 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='scaenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 8 (Snom)'), (SELECT param_id FROM provisioning_params WHERE name='extension'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 8 (Snom)'), (SELECT param_id FROM provisioning_params WHERE name='password'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 8 (Snom)'), (SELECT param_id FROM provisioning_params WHERE name='callerid'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 9 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='extension'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 9 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='password'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 9 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardnoanswer'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 9 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardnoanswerenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 9 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardbusy'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 9 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardbusyenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 9 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardall'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 9 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardallenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 9 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='callerid'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 9 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='polycomringtone'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 9 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='callwaitingenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 9 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='scaenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 9 (Snom)'), (SELECT param_id FROM provisioning_params WHERE name='extension'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 9 (Snom)'), (SELECT param_id FROM provisioning_params WHERE name='password'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 9 (Snom)'), (SELECT param_id FROM provisioning_params WHERE name='callerid'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 10 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='extension'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 10 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='password'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 10 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardnoanswer'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 10 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardnoanswerenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 10 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardbusy'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 10 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardbusyenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 10 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardall'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 10 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardallenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 10 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='callerid'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 10 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='polycomringtone'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 10 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='callwaitingenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 10 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='scaenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 10 (Snom)'), (SELECT param_id FROM provisioning_params WHERE name='extension'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 10 (Snom)'), (SELECT param_id FROM provisioning_params WHERE name='password'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 10 (Snom)'), (SELECT param_id FROM provisioning_params WHERE name='callerid'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 11 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='extension'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 11 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='password'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 11 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardnoanswer'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 11 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardnoanswerenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 11 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardbusy'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 11 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardbusyenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 11 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardall'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 11 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardallenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 11 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='callerid'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 11 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='polycomringtone'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 11 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='callwaitingenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 11 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='scaenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 11 (Snom)'), (SELECT param_id FROM provisioning_params WHERE name='extension'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 11 (Snom)'), (SELECT param_id FROM provisioning_params WHERE name='password'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 11 (Snom)'), (SELECT param_id FROM provisioning_params WHERE name='callerid'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 12 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='extension'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 12 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='password'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 12 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardnoanswer'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 12 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardnoanswerenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 12 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardbusy'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 12 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardbusyenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 12 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardall'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 12 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardallenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 12 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='callerid'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 12 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='polycomringtone'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 12 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='callwaitingenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 12 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='scaenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 12 (Snom)'), (SELECT param_id FROM provisioning_params WHERE name='extension'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 12 (Snom)'), (SELECT param_id FROM provisioning_params WHERE name='password'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 12 (Snom)'), (SELECT param_id FROM provisioning_params WHERE name='callerid'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 13 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='extension'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 13 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='password'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 13 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardnoanswer'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 13 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardnoanswerenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 13 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardbusy'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 13 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardbusyenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 13 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardall'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 13 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardallenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 13 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='callerid'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 13 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='polycomringtone'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 13 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='callwaitingenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 13 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='scaenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 14 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='extension'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 14 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='password'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 14 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardnoanswer'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 14 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardnoanswerenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 14 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardbusy'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 14 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardbusyenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 14 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardall'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 14 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardallenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 14 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='callerid'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 14 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='polycomringtone'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 14 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='callwaitingenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 14 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='scaenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 15 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='extension'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 15 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='password'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 15 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardnoanswer'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 15 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardnoanswerenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 15 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardbusy'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 15 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardbusyenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 15 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardall'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 15 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardallenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 15 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='callerid'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 15 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='polycomringtone'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 15 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='callwaitingenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 15 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='scaenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 16 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='extension'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 16 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='password'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 16 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardnoanswer'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 16 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardnoanswerenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 16 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardbusy'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 16 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardbusyenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 16 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardall'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 16 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardallenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 16 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='callerid'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 16 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='polycomringtone'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 16 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='callwaitingenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 16 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='scaenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 17 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='extension'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 17 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='password'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 17 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardnoanswer'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 17 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardnoanswerenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 17 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardbusy'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 17 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardbusyenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 17 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardall'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 17 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardallenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 17 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='callerid'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 17 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='polycomringtone'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 17 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='callwaitingenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 17 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='scaenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 18 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='extension'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 18 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='password'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 18 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardnoanswer'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 18 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardnoanswerenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 18 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardbusy'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 18 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardbusyenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 18 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardall'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 18 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardallenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 18 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='callerid'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 18 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='polycomringtone'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 18 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='callwaitingenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 18 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='scaenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 19 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='extension'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 19 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='password'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 19 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardnoanswer'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 19 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardnoanswerenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 19 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardbusy'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 19 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardbusyenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 19 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardall'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 19 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardallenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 19 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='callerid'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 19 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='polycomringtone'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 19 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='callwaitingenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 19 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='scaenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 20 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='extension'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 20 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='password'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 20 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardnoanswer'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 20 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardnoanswerenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 20 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardbusy'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 20 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardbusyenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 20 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardall'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 20 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardallenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 20 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='callerid'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 20 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='polycomringtone'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 20 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='callwaitingenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 20 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='scaenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 21 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='extension'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 21 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='password'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 21 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardnoanswer'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 21 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardnoanswerenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 21 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardbusy'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 21 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardbusyenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 21 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardall'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 21 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardallenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 21 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='callerid'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 21 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='polycomringtone'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 21 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='callwaitingenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 21 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='scaenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 22 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='extension'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 22 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='password'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 22 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardnoanswer'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 22 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardnoanswerenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 22 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardbusy'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 22 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardbusyenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 22 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardall'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 22 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardallenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 22 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='callerid'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 22 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='polycomringtone'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 22 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='callwaitingenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 22 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='scaenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 23 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='extension'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 23 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='password'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 23 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardnoanswer'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 23 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardnoanswerenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 23 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardbusy'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 23 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardbusyenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 23 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardall'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 23 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardallenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 23 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='callerid'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 23 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='polycomringtone'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 23 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='callwaitingenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 23 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='scaenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 24 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='extension'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 24 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='password'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 24 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardnoanswer'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 24 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardnoanswerenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 24 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardbusy'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 24 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardbusyenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 24 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardall'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 24 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardallenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 24 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='callerid'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 24 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='polycomringtone'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 24 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='callwaitingenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 24 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='scaenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 25 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='extension'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 25 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='password'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 25 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardnoanswer'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 25 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardnoanswerenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 25 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardbusy'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 25 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardbusyenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 25 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardall'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 25 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardallenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 25 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='callerid'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 25 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='polycomringtone'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 25 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='callwaitingenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 25 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='scaenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 26 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='extension'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 26 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='password'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 26 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardnoanswer'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 26 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardnoanswerenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 26 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardbusy'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 26 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardbusyenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 26 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardall'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 26 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardallenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 26 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='callerid'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 26 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='polycomringtone'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 26 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='callwaitingenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 26 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='scaenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 27 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='extension'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 27 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='password'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 27 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardnoanswer'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 27 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardnoanswerenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 27 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardbusy'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 27 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardbusyenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 27 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardall'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 27 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardallenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 27 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='callerid'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 27 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='polycomringtone'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 27 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='callwaitingenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 27 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='scaenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 28 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='extension'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 28 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='password'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 28 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardnoanswer'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 28 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardnoanswerenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 28 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardbusy'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 28 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardbusyenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 28 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardall'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 28 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardallenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 28 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='callerid'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 28 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='polycomringtone'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 28 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='callwaitingenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 28 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='scaenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 29 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='extension'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 29 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='password'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 29 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardnoanswer'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 29 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardnoanswerenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 29 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardbusy'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 29 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardbusyenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 29 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardall'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 29 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardallenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 29 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='callerid'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 29 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='polycomringtone'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 29 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='callwaitingenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 29 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='scaenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 30 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='extension'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 30 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='password'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 30 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardnoanswer'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 30 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardnoanswerenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 30 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardbusy'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 30 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardbusyenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 30 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardall'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 30 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardallenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 30 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='callerid'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 30 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='polycomringtone'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 30 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='callwaitingenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 30 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='scaenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 31 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='extension'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 31 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='password'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 31 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardnoanswer'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 31 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardnoanswerenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 31 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardbusy'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 31 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardbusyenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 31 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardall'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 31 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardallenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 31 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='callerid'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 31 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='polycomringtone'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 31 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='callwaitingenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 31 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='scaenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 32 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='extension'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 32 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='password'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 32 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardnoanswer'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 32 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardnoanswerenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 32 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardbusy'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 32 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardbusyenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 32 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardall'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 32 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardallenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 32 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='callerid'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 32 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='polycomringtone'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 32 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='callwaitingenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 32 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='scaenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 33 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='extension'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 33 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='password'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 33 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardnoanswer'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 33 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardnoanswerenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 33 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardbusy'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 33 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardbusyenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 33 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardall'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 33 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardallenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 33 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='callerid'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 33 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='polycomringtone'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 33 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='callwaitingenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 33 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='scaenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 34 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='extension'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 34 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='password'), null, false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 34 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardnoanswer'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 34 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardnoanswerenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 34 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardbusy'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 34 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardbusyenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 34 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardall'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 34 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='forwardallenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 34 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='callerid'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 34 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='polycomringtone'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 34 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='callwaitingenable'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Registration 34 (Polycom)'), (SELECT param_id FROM provisioning_params WHERE name='scaenable'), null, true)",
	-- Services
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Office Phone'), (SELECT param_id FROM provisioning_params WHERE name='forwarding'), 'true', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Office Phone'), (SELECT param_id FROM provisioning_params WHERE name='hotlineenable'), 'false', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Office Phone'), (SELECT param_id FROM provisioning_params WHERE name='callhistoryenable'), 'true', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Office Phone'), (SELECT param_id FROM provisioning_params WHERE name='missedcallhistoryenable'), 'true', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Office Phone'), (SELECT param_id FROM provisioning_params WHERE name='receivedcallhistoryenable'), 'true', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Office Phone'), (SELECT param_id FROM provisioning_params WHERE name='callwaitingenable'), 'true', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Office Phone'), (SELECT param_id FROM provisioning_params WHERE name='speeddialenable'), 'true', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Office Phone'), (SELECT param_id FROM provisioning_params WHERE name='mailbox'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Office Phone'), (SELECT param_id FROM provisioning_params WHERE name='mailcallback'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Office Phone'), (SELECT param_id FROM provisioning_params WHERE name='provpollingtime'), '02:00', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Office Phone'), (SELECT param_id FROM provisioning_params WHERE name='dndenable'), 'true', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Office Phone'), (SELECT param_id FROM provisioning_params WHERE name='softkey1'), 'false', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Office Phone'), (SELECT param_id FROM provisioning_params WHERE name='softkey2'), 'false', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Office Phone'), (SELECT param_id FROM provisioning_params WHERE name='softkey3'), 'false', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Office Phone'), (SELECT param_id FROM provisioning_params WHERE name='softkey4'), 'false', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Office Phone'), (SELECT param_id FROM provisioning_params WHERE name='softkey5'), 'false', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Office Phone'), (SELECT param_id FROM provisioning_params WHERE name='softkey6'), 'false', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Office Phone'), (SELECT param_id FROM provisioning_params WHERE name='softkey7'), 'false', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Office Phone'), (SELECT param_id FROM provisioning_params WHERE name='softkey8'), 'false', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Office Phone'), (SELECT param_id FROM provisioning_params WHERE name='softkey9'), 'false', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Office Phone'), (SELECT param_id FROM provisioning_params WHERE name='softkey10'), 'false', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Residential Phone'), (SELECT param_id FROM provisioning_params WHERE name='forwarding'), 'true', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Residential Phone'), (SELECT param_id FROM provisioning_params WHERE name='hotlineenable'), 'false', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Residential Phone'), (SELECT param_id FROM provisioning_params WHERE name='callhistoryenable'), 'true', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Residential Phone'), (SELECT param_id FROM provisioning_params WHERE name='missedcallhistoryenable'), 'true', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Residential Phone'), (SELECT param_id FROM provisioning_params WHERE name='receivedcallhistoryenable'), 'true', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Residential Phone'), (SELECT param_id FROM provisioning_params WHERE name='callwaitingenable'), 'true', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Residential Phone'), (SELECT param_id FROM provisioning_params WHERE name='speeddialenable'), 'true', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Residential Phone'), (SELECT param_id FROM provisioning_params WHERE name='mailbox'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Residential Phone'), (SELECT param_id FROM provisioning_params WHERE name='mailcallback'), null, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Residential Phone'), (SELECT param_id FROM provisioning_params WHERE name='provpollingtime'), '10:00', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Residential Phone'), (SELECT param_id FROM provisioning_params WHERE name='dndenable'), 'true', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Residential Phone'), (SELECT param_id FROM provisioning_params WHERE name='softkey1'), 'false', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Residential Phone'), (SELECT param_id FROM provisioning_params WHERE name='softkey2'), 'false', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Residential Phone'), (SELECT param_id FROM provisioning_params WHERE name='softkey3'), 'false', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Residential Phone'), (SELECT param_id FROM provisioning_params WHERE name='softkey4'), 'false', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Residential Phone'), (SELECT param_id FROM provisioning_params WHERE name='softkey5'), 'false', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Residential Phone'), (SELECT param_id FROM provisioning_params WHERE name='softkey6'), 'false', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Residential Phone'), (SELECT param_id FROM provisioning_params WHERE name='softkey7'), 'false', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Residential Phone'), (SELECT param_id FROM provisioning_params WHERE name='softkey8'), 'false', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Residential Phone'), (SELECT param_id FROM provisioning_params WHERE name='softkey9'), 'false', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Residential Phone'), (SELECT param_id FROM provisioning_params WHERE name='softkey10'), 'false', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Public Phone'), (SELECT param_id FROM provisioning_params WHERE name='forwarding'), 'false', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Public Phone'), (SELECT param_id FROM provisioning_params WHERE name='hotlineenable'), 'false', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Public Phone'), (SELECT param_id FROM provisioning_params WHERE name='callhistoryenable'), 'false', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Public Phone'), (SELECT param_id FROM provisioning_params WHERE name='missedcallhistoryenable'), 'true', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Public Phone'), (SELECT param_id FROM provisioning_params WHERE name='receivedcallhistoryenable'), 'true', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Public Phone'), (SELECT param_id FROM provisioning_params WHERE name='callwaitingenable'), 'true', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Public Phone'), (SELECT param_id FROM provisioning_params WHERE name='speeddialenable'), 'false', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Public Phone'), (SELECT param_id FROM provisioning_params WHERE name='provpollingtime'), '02:00', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Public Phone'), (SELECT param_id FROM provisioning_params WHERE name='dndenable'), 'false', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Public Phone'), (SELECT param_id FROM provisioning_params WHERE name='softkey1'), 'false', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Public Phone'), (SELECT param_id FROM provisioning_params WHERE name='softkey2'), 'false', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Public Phone'), (SELECT param_id FROM provisioning_params WHERE name='softkey3'), 'false', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Public Phone'), (SELECT param_id FROM provisioning_params WHERE name='softkey4'), 'false', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Public Phone'), (SELECT param_id FROM provisioning_params WHERE name='softkey5'), 'false', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Public Phone'), (SELECT param_id FROM provisioning_params WHERE name='softkey6'), 'false', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Public Phone'), (SELECT param_id FROM provisioning_params WHERE name='softkey7'), 'false', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Public Phone'), (SELECT param_id FROM provisioning_params WHERE name='softkey8'), 'false', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Public Phone'), (SELECT param_id FROM provisioning_params WHERE name='softkey9'), 'false', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Public Phone'), (SELECT param_id FROM provisioning_params WHERE name='softkey10'), 'false', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Hotline'), (SELECT param_id FROM provisioning_params WHERE name='forwarding'), 'false', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Hotline'), (SELECT param_id FROM provisioning_params WHERE name='hotlineenable'), 'true', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Hotline'), (SELECT param_id FROM provisioning_params WHERE name='hotlinedestination'), '', true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Hotline'), (SELECT param_id FROM provisioning_params WHERE name='callhistoryenable'), 'false', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Hotline'), (SELECT param_id FROM provisioning_params WHERE name='missedcallhistoryenable'), 'false', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Hotline'), (SELECT param_id FROM provisioning_params WHERE name='receivedcallhistoryenable'), 'false', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Hotline'), (SELECT param_id FROM provisioning_params WHERE name='callwaitingenable'), 'true', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Hotline'), (SELECT param_id FROM provisioning_params WHERE name='speeddialenable'), 'false', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Hotline'), (SELECT param_id FROM provisioning_params WHERE name='provpollingtime'), '02:00', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Hotline'), (SELECT param_id FROM provisioning_params WHERE name='dndenable'), 'false', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Hotline'), (SELECT param_id FROM provisioning_params WHERE name='softkey1'), 'false', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Hotline'), (SELECT param_id FROM provisioning_params WHERE name='softkey2'), 'false', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Hotline'), (SELECT param_id FROM provisioning_params WHERE name='softkey3'), 'false', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Hotline'), (SELECT param_id FROM provisioning_params WHERE name='softkey4'), 'false', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Hotline'), (SELECT param_id FROM provisioning_params WHERE name='softkey5'), 'false', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Hotline'), (SELECT param_id FROM provisioning_params WHERE name='softkey6'), 'false', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Hotline'), (SELECT param_id FROM provisioning_params WHERE name='softkey7'), 'false', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Hotline'), (SELECT param_id FROM provisioning_params WHERE name='softkey8'), 'false', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Hotline'), (SELECT param_id FROM provisioning_params WHERE name='softkey9'), 'false', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Hotline'), (SELECT param_id FROM provisioning_params WHERE name='softkey10'), 'false', false)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Zone 00 (CyberData Paging Zone Controller)'), (SELECT param_id FROM provisioning_params WHERE name='pagingport1'), true, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Zone 00 (CyberData Paging Zone Controller)'), (SELECT param_id FROM provisioning_params WHERE name='pagingport2'), true, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Zone 00 (CyberData Paging Zone Controller)'), (SELECT param_id FROM provisioning_params WHERE name='pagingport3'), true, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Zone 00 (CyberData Paging Zone Controller)'), (SELECT param_id FROM provisioning_params WHERE name='pagingport4'), true, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Zone 01 (CyberData Paging Zone Controller)'), (SELECT param_id FROM provisioning_params WHERE name='pagingport1'), true, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Zone 01 (CyberData Paging Zone Controller)'), (SELECT param_id FROM provisioning_params WHERE name='pagingport2'), false, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Zone 01 (CyberData Paging Zone Controller)'), (SELECT param_id FROM provisioning_params WHERE name='pagingport3'), false, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Zone 01 (CyberData Paging Zone Controller)'), (SELECT param_id FROM provisioning_params WHERE name='pagingport4'), false, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Zone 02 (CyberData Paging Zone Controller)'), (SELECT param_id FROM provisioning_params WHERE name='pagingport1'), false, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Zone 02 (CyberData Paging Zone Controller)'), (SELECT param_id FROM provisioning_params WHERE name='pagingport2'), true, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Zone 02 (CyberData Paging Zone Controller)'), (SELECT param_id FROM provisioning_params WHERE name='pagingport3'), false, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Zone 02 (CyberData Paging Zone Controller)'), (SELECT param_id FROM provisioning_params WHERE name='pagingport4'), false, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Zone 03 (CyberData Paging Zone Controller)'), (SELECT param_id FROM provisioning_params WHERE name='pagingport1'), false, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Zone 03 (CyberData Paging Zone Controller)'), (SELECT param_id FROM provisioning_params WHERE name='pagingport2'), false, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Zone 03 (CyberData Paging Zone Controller)'), (SELECT param_id FROM provisioning_params WHERE name='pagingport3'), true, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Zone 03 (CyberData Paging Zone Controller)'), (SELECT param_id FROM provisioning_params WHERE name='pagingport4'), false, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Zone 04 (CyberData Paging Zone Controller)'), (SELECT param_id FROM provisioning_params WHERE name='pagingport1'), false, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Zone 04 (CyberData Paging Zone Controller)'), (SELECT param_id FROM provisioning_params WHERE name='pagingport2'), false, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Zone 04 (CyberData Paging Zone Controller)'), (SELECT param_id FROM provisioning_params WHERE name='pagingport3'), false, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Zone 04 (CyberData Paging Zone Controller)'), (SELECT param_id FROM provisioning_params WHERE name='pagingport4'), true, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Zone 05 (CyberData Paging Zone Controller)'), (SELECT param_id FROM provisioning_params WHERE name='pagingport1'), true, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Zone 05 (CyberData Paging Zone Controller)'), (SELECT param_id FROM provisioning_params WHERE name='pagingport2'), true, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Zone 05 (CyberData Paging Zone Controller)'), (SELECT param_id FROM provisioning_params WHERE name='pagingport3'), false, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Zone 05 (CyberData Paging Zone Controller)'), (SELECT param_id FROM provisioning_params WHERE name='pagingport4'), false, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Zone 06 (CyberData Paging Zone Controller)'), (SELECT param_id FROM provisioning_params WHERE name='pagingport1'), true, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Zone 06 (CyberData Paging Zone Controller)'), (SELECT param_id FROM provisioning_params WHERE name='pagingport2'), false, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Zone 06 (CyberData Paging Zone Controller)'), (SELECT param_id FROM provisioning_params WHERE name='pagingport3'), true, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Zone 06 (CyberData Paging Zone Controller)'), (SELECT param_id FROM provisioning_params WHERE name='pagingport4'), false, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Zone 07 (CyberData Paging Zone Controller)'), (SELECT param_id FROM provisioning_params WHERE name='pagingport1'), false, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Zone 07 (CyberData Paging Zone Controller)'), (SELECT param_id FROM provisioning_params WHERE name='pagingport2'), true, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Zone 07 (CyberData Paging Zone Controller)'), (SELECT param_id FROM provisioning_params WHERE name='pagingport3'), true, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Zone 07 (CyberData Paging Zone Controller)'), (SELECT param_id FROM provisioning_params WHERE name='pagingport4'), false, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Zone 08 (CyberData Paging Zone Controller)'), (SELECT param_id FROM provisioning_params WHERE name='pagingport1'), true, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Zone 08 (CyberData Paging Zone Controller)'), (SELECT param_id FROM provisioning_params WHERE name='pagingport2'), true, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Zone 08 (CyberData Paging Zone Controller)'), (SELECT param_id FROM provisioning_params WHERE name='pagingport3'), true, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Zone 08 (CyberData Paging Zone Controller)'), (SELECT param_id FROM provisioning_params WHERE name='pagingport4'), false, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Zone 09 (CyberData Paging Zone Controller)'), (SELECT param_id FROM provisioning_params WHERE name='pagingport1'), true, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Zone 09 (CyberData Paging Zone Controller)'), (SELECT param_id FROM provisioning_params WHERE name='pagingport2'), false, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Zone 09 (CyberData Paging Zone Controller)'), (SELECT param_id FROM provisioning_params WHERE name='pagingport3'), false, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Zone 09 (CyberData Paging Zone Controller)'), (SELECT param_id FROM provisioning_params WHERE name='pagingport4'), true, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Zone 10 (CyberData Paging Zone Controller)'), (SELECT param_id FROM provisioning_params WHERE name='pagingport1'), false, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Zone 10 (CyberData Paging Zone Controller)'), (SELECT param_id FROM provisioning_params WHERE name='pagingport2'), true, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Zone 10 (CyberData Paging Zone Controller)'), (SELECT param_id FROM provisioning_params WHERE name='pagingport3'), false, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Zone 10 (CyberData Paging Zone Controller)'), (SELECT param_id FROM provisioning_params WHERE name='pagingport4'), true, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Zone 11 (CyberData Paging Zone Controller)'), (SELECT param_id FROM provisioning_params WHERE name='pagingport1'), true, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Zone 11 (CyberData Paging Zone Controller)'), (SELECT param_id FROM provisioning_params WHERE name='pagingport2'), true, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Zone 11 (CyberData Paging Zone Controller)'), (SELECT param_id FROM provisioning_params WHERE name='pagingport3'), false, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Zone 11 (CyberData Paging Zone Controller)'), (SELECT param_id FROM provisioning_params WHERE name='pagingport4'), true, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Zone 12 (CyberData Paging Zone Controller)'), (SELECT param_id FROM provisioning_params WHERE name='pagingport1'), false, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Zone 12 (CyberData Paging Zone Controller)'), (SELECT param_id FROM provisioning_params WHERE name='pagingport2'), false, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Zone 12 (CyberData Paging Zone Controller)'), (SELECT param_id FROM provisioning_params WHERE name='pagingport3'), true, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Zone 12 (CyberData Paging Zone Controller)'), (SELECT param_id FROM provisioning_params WHERE name='pagingport4'), true, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Zone 13 (CyberData Paging Zone Controller)'), (SELECT param_id FROM provisioning_params WHERE name='pagingport1'), true, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Zone 13 (CyberData Paging Zone Controller)'), (SELECT param_id FROM provisioning_params WHERE name='pagingport2'), false, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Zone 13 (CyberData Paging Zone Controller)'), (SELECT param_id FROM provisioning_params WHERE name='pagingport3'), true, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Zone 13 (CyberData Paging Zone Controller)'), (SELECT param_id FROM provisioning_params WHERE name='pagingport4'), true, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Zone 14 (CyberData Paging Zone Controller)'), (SELECT param_id FROM provisioning_params WHERE name='pagingport1'), false, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Zone 14 (CyberData Paging Zone Controller)'), (SELECT param_id FROM provisioning_params WHERE name='pagingport2'), true, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Zone 14 (CyberData Paging Zone Controller)'), (SELECT param_id FROM provisioning_params WHERE name='pagingport3'), true, true)",
	"INSERT INTO param_groups_to_params VALUES((SELECT group_id FROM provisioning_groups WHERE label='Zone 14 (CyberData Paging Zone Controller)'), (SELECT param_id FROM provisioning_params WHERE name='pagingport4'), true, true)",
}

-- List of each parameter used in any way for any device - mostly for how to display
mymodule.provisioning_params = {
	"CREATE TABLE provisioning_params (param_id SERIAL PRIMARY KEY, name VARCHAR(255) UNIQUE, type VARCHAR(255), label VARCHAR(255), descr VARCHAR(255), value VARCHAR(255), seq INTEGER, regexp VARCHAR(255), validate text)",
	"CREATE INDEX params_name_idx ON provisioning_params (name)",
	"INSERT INTO provisioning_params VALUES(default, 'mac', 'text', 'MAC Address', 'Capitalized hex digits with no punctuation', '', '1', '^%x%x%x%x%x%x%x%x%x%x%x%x$', E'local value, functions, params = ...\nvalue = string.upper(value)\nlocal others = functions.getselectresponse(\"SELECT count(*) FROM provisioning_values WHERE param_id=\\'\"..params.value.device.value.mac.param_id..\"\\' AND device_id!=\\'\"..params.value.device_id.value..\"\\' AND value=\\'\"..value..\"\\'\")\nif tonumber(others[1].count) > 0 then\n\treturn value, \"MAC Address must be unique\"\nend\nreturn value')",
	"INSERT INTO provisioning_params VALUES(default, 'template', 'select', 'Template', '', '', '2', '', null)",
	"INSERT INTO provisioning_params VALUES(default, 'registrar', 'text', 'SIP Registrar', '', '', '3', '', null)",
	"INSERT INTO provisioning_params VALUES(default, 'digitmap', 'text', 'Digit Map', 'Phone dial pattern based on section 2.1.5 of RFC 3435, plus a comma to turn dialtone back on', '', '4', '^[*#0-9xT|,.%[%]-]*$', null)",
	"INSERT INTO provisioning_params VALUES(default, 'digitmaptimeout', 'text', 'Digit Map Timeout', E'Timeout in seconds for each segment of digit map (separated by \\'|\\')', '', '5', '^[0-9|]*$', null)",
	"INSERT INTO provisioning_params VALUES(default, 'sntpserver', 'text', 'SNTP Server', '', '', '6', '', null)",
	"INSERT INTO provisioning_params VALUES(default, 'timezone', 'text', 'Posix Time Zone', 'See the expanded form (no colon) of TZ variable at http://pubs.opengroup.org/onlinepubs/9699919799/basedefs/V1_chap08.html', 'UTC0', '7', '^%a+[+-]?%d+', null)",
	"INSERT INTO provisioning_params VALUES(default, 'musiconhold', 'text', 'Music-on-hold URI', '', '', '8', '', null)",
	"INSERT INTO provisioning_params VALUES(default, 'homepage', 'text', 'Browser Homepage', '', '', '9', '', null)",
	"INSERT INTO provisioning_params VALUES(default, 'adminpassword', 'text', 'Administration Password', '', '', '10', '', null)",
	"INSERT INTO provisioning_params VALUES(default, 'pcportenable', 'boolean', 'PC Ethernet Port Enable', '', 'true', '11', '', null)",
	"INSERT INTO provisioning_params VALUES(default, 'urldialingenable', 'boolean', 'URL Dialing Enable', '', 'false', '12', '', null)",
	"INSERT INTO provisioning_params VALUES(default, 'provpollingtime', 'select', 'Provisioning polling start time', 'The provisioning server is automatically polled for upgrades at random between the start time and one hour later', '02:00', '13', '', null)",
	"INSERT INTO provisioning_params VALUES(default, 'militarytimeenable', 'boolean', '24-hour Clock Mode Enable', '', 'false', '14', '', null)",
	"INSERT INTO provisioning_params VALUES(default, 'clockenable', 'boolean', 'Clock Enable', '', 'true', '14', '', null)",
	"INSERT INTO provisioning_params VALUES(default, 'displaylang', 'select', 'Display screen language', 'Check if your phone supports the selected language', '', '15', '', null)",
	"INSERT INTO provisioning_params VALUES(default, 'algoringvol', 'select', 'Ring Volume (Algo)', '', '5', '16', '', null)",
	"INSERT INTO provisioning_params VALUES(default, 'algopagevol', 'select', 'Page Volume (Algo)', '', '5', '17', '', null)",
	"INSERT INTO provisioning_params values(default, 'algospkmode', 'select', 'Speaker Mode (Algo)', '', '0', '18', '', null)",
	"INSERT INTO provisioning_params values(default, 'cyberspeakervolume', 'select', 'Speaker Volume (CyberData)', '', '4', '19', '', null)",
	"INSERT INTO provisioning_params values(default, 'cyberringvolume', 'select', 'Ring Volume (CyberData)', '', '4', '20', '', null)",
	"INSERT INTO provisioning_params values(default, 'cybermicgain', 'select', 'Microphone Gain (CyberData)', '', '4', '21', '', null)",
	"INSERT INTO provisioning_params values(default, 'activaterelaywithdtmf', 'boolean', 'Activate Relay with DTMF Enable', '', 'true', '22', '', null)",
	"INSERT INTO provisioning_params values(default, 'dtmfactivationcode', 'text', 'Relay DTMF Activation Code', '', '321', '23', '^[*#0-9]*$', E'local value, functions, params = ...\nif value ~= \"\" then\nlocal len = string.len(value)\n if len >= 26 then\n  return value, \"Maximum 25 digits\"\n end\nend\nreturn value')",
	"INSERT INTO provisioning_params values(default, 'dtmfactivationduration', 'text', 'Relay DTMF Activation Duration', 'Value in seconds', '2', '24', '^%d*$', E'local value, functions, params = ...\nif value ~= \"\" then\nlocal len = string.len(value)\n if len >= 6 then\n  return value, \"Maximum 5 digits\"\n end\nend\nreturn value')",
	"INSERT INTO provisioning_params values(default, 'playtonewhilerelayactive', 'boolean', 'Relay Activation Tone Enable', '', 'true', '25', '', null)",
	"INSERT INTO provisioning_params values(default, 'activaterelayduringring', 'boolean', 'Activate Relay During Ring Enable', '', 'false', '26', '', null)",
	"INSERT INTO provisioning_params values(default, 'activaterelayduringcall', 'boolean', 'Activate Relay While Call Active Enable', '', 'false', '27', '', null)",
	"INSERT INTO provisioning_params values(default, 'bypassdtmf', 'boolean', 'Bypass SIP DTMF Entry', 'Bypassing DTMF will result in all SIP calls being played to Zone 0', 'false', '28', '', null)",
	"INSERT INTO provisioning_params values(default, 'beeponinitialization', 'boolean', 'Beep On Initialization', '', 'true', '29', '', null)",
	"INSERT INTO provisioning_params values(default, 'beepbeforepage', 'boolean', 'Beep on page', '', 'true', '30', '', null)",
	"INSERT INTO provisioning_params values(default, 'pagingport1', 'boolean', 'Paging Port 1 Enable', '', 'true', '31', '', null)",
	"INSERT INTO provisioning_params values(default, 'pagingport2', 'boolean', 'Paging Port 2 Enable', '', 'true', '32', '', null)",
	"INSERT INTO provisioning_params values(default, 'pagingport3', 'boolean', 'Paging Port 3 Enable', '', 'true', '33', '', null)",
	"INSERT INTO provisioning_params values(default, 'pagingport4', 'boolean', 'Paging Port 4 Enable', '', 'true', '34', '', null)",
	"INSERT INTO provisioning_params values(default, 'autoanswerincomingcalls', 'boolean', 'Auto Answer Incoming Calls Enable', '', 'true', '35', '', null)",
	"INSERT INTO provisioning_params values(default, 'sensornormallyclosed', 'boolean', 'Door Sensor Normally Closed Enable', '', 'false', '36', '', null)",
	"INSERT INTO provisioning_params values(default, 'sensortimeout', 'text', 'Door Sensor Open Timeout', 'Value in seconds', '0', '37', '^%d*$', E'local value, functions, params = ...\nif value == \\'\\' then\n return value, \\'Empty not accepted\\'\nend\nlocal value = tonumber(value)\nif value == nil then\n return value, \\'Only numbers are allowed\\'\nend\nif value >= 99999 then\n return value, \\'Maximum 99999\\'\nend\nreturn value')",
	"INSERT INTO provisioning_params values(default, 'sensorflashled', 'boolean', 'Door Sensor Flash Button LED Enable', '', 'false', '38', '', null)",
	"INSERT INTO provisioning_params values(default, 'sensoractivaterelay', 'boolean', 'Door Sensor Activate Relay Enable', '', 'false', '39', '', null)",
	"INSERT INTO provisioning_params values(default, 'sensorplaylocally', 'boolean', 'Door Sensor Play Audio Locally Enable', '', 'false', '40', '', null)",
	"INSERT INTO provisioning_params values(default, 'sensorcall', 'boolean', 'Door Sensor Call to Extension Enable', '', 'false', '41', '', null)",
	"INSERT INTO provisioning_params values(default, 'sensordialoutextension', 'text', 'Door Sensor Dial Out Extension', '', '', '42', '^[*#0-9]*$', E'local value, functions, params = ...\nif value ~= \\'\\' then\nlocal len = string.len(value)\n if len >= 65 then\n  return value, \\'Maximum 64 digits\\'\n end\nend\nreturn value')",
	"INSERT INTO provisioning_params values(default, 'sensordialoutid', 'text', 'Door Sensor Dial Out ID', '', '', '43', '', E'local value, functions, params = ...\nif value ~= \\'\\' then\nlocal len = string.len(value)\n if len >= 65 then\n  return value, \\'Maximum 64 characters\\'\n end\nend\nreturn value')",
	"INSERT INTO provisioning_params values(default, 'sensorrepeat', 'text', 'Door Sensor Repeat Message', '', '1', '44', '', E'local value, functions, params = ...\nif value == \\'\\' then\n return value, \\'Empty not accepted\\'\nend\nlocal value = tonumber(value)\nif value == nil then\n return value, \\'Only numbers are allowed\\'\nend\nif value >= 65537 then\n return value, \\'Maximum 65536\\'\nend\nreturn value')",
	"INSERT INTO provisioning_params values(default, 'sensorplayremotely', 'boolean', 'Door Sensor Play Audio Enable', '', 'false', '45', '', null)",
	"INSERT INTO provisioning_params VALUES(default, 'dectenable', 'boolean', 'VVX D60 Enable', '', 'false', '46', '', null)",
	"INSERT INTO provisioning_params VALUES(default, 'extension', 'text', 'Extension', '', '', '101', '^%d*$', null)",
	"INSERT INTO provisioning_params VALUES(default, 'password', 'text', 'Password', '', '', '102', '', null)",
	"INSERT INTO provisioning_params VALUES(default, 'forwardnoanswerenable', 'boolean', 'Forward on No-answer Enable', '', 'false', '103', '', null)",
	"INSERT INTO provisioning_params VALUES(default, 'forwardnoanswer', 'text', 'Forward on No-answer Destination', 'Callers will be transferred to this extension when you don''t answer after 20 seconds', '', '104', '', null)",
	"INSERT INTO provisioning_params VALUES(default, 'forwardbusyenable', 'boolean', 'Forward on Busy Enable', '', 'false', '105', '', null)",
	"INSERT INTO provisioning_params VALUES(default, 'forwardbusy', 'text', 'Forward on Busy Destination', 'Callers will be transferred to this extension when the line is busy', '', '106', '', null)",
	"INSERT INTO provisioning_params VALUES(default, 'forwardallenable', 'boolean', 'Forward All Calls Enable', '', 'false', '107', '', null)",
	"INSERT INTO provisioning_params VALUES(default, 'forwardall', 'text', 'Forward All Calls Destination', 'All calls will be transferred to this extension', '', '108', '', null)",
	"INSERT INTO provisioning_params VALUES(default, 'callerid', 'text', 'Caller ID String', '', '', '109', '', null)",
	"INSERT INTO provisioning_params VALUES(default, 'polycomringtone', 'select', 'Ring Tone (Polycom)', '', '2', '110', '', null)",
	"INSERT INTO provisioning_params VALUES(default, 'algoringtone', 'select', 'Ring Tone (Algo)', '', 'warble2-med.wav', '111', '', null)",
	"INSERT INTO provisioning_params VALUES(default, 'algopagetone', 'select', 'Page Tone (Algo)', '', 'warble2-med.wav', '112', '', null)",
	"INSERT INTO provisioning_params VALUES(default, 'scaenable', 'boolean', 'Shared Call Appearance Enable', '', 'false', '120', '', null)",
	"INSERT INTO provisioning_params VALUES(default, 'forwarding', 'boolean', 'Forwarding Enable', '', 'true', '201', '', null)",
	"INSERT INTO provisioning_params VALUES(default, 'hotlineenable', 'boolean', 'Hotline Enable', '', 'false', '202', '', null)",
	"INSERT INTO provisioning_params VALUES(default, 'hotlinedestination', 'text', 'Hotline Destination', '', '', '203', '', null)",
	"INSERT INTO provisioning_params VALUES(default, 'callhistoryenable', 'boolean', 'Placed Call History Enable', '', 'true', '204', '', null)",
	"INSERT INTO provisioning_params VALUES(default, 'missedcallhistoryenable', 'boolean', 'Missed Call History Enable', '', 'true', '204', '', null)",
	"INSERT INTO provisioning_params VALUES(default, 'receivedcallhistoryenable', 'boolean', 'Received Call History Enable', '', 'true', '204', '', null)",
	"INSERT INTO provisioning_params VALUES(default, 'callwaitingenable', 'boolean', 'Call Waiting Enable', '', 'true', '205', '', null)",
	"INSERT INTO provisioning_params VALUES(default, 'speeddialenable', 'boolean', 'Speed Dial Enable', '', 'true', '206', '', null)",
	"INSERT INTO provisioning_params VALUES(default, 'mailbox', 'text', 'Voice Mailbox', 'Mailbox extension or URL', '', '207', '', null)",
	"INSERT INTO provisioning_params VALUES(default, 'mailcallback', 'text', 'Voice Mailbox Callback', 'Extension or URL for mailbox message retrieval', '', '208', '', null)",
	"INSERT INTO provisioning_params VALUES(default, 'dndenable', 'boolean', 'Do Not Disturb Enable', '', 'true', '220', '', null)",
	"INSERT INTO provisioning_params VALUES(default, 'softkey1', 'boolean', 'Soft Key 1 Enable', '', 'false', '301', '', null)",
	"INSERT INTO provisioning_params VALUES(default, 'softkey2', 'boolean', 'Soft Key 2 Enable', '', 'false', '302', '', null)",
	"INSERT INTO provisioning_params VALUES(default, 'softkey3', 'boolean', 'Soft Key 3 Enable', '', 'false', '303', '', null)",
	"INSERT INTO provisioning_params VALUES(default, 'softkey4', 'boolean', 'Soft Key 4 Enable', '', 'false', '304', '', null)",
	"INSERT INTO provisioning_params VALUES(default, 'softkey5', 'boolean', 'Soft Key 5 Enable', '', 'false', '305', '', null)",
	"INSERT INTO provisioning_params VALUES(default, 'softkey6', 'boolean', 'Soft Key 6 Enable', '', 'false', '306', '', null)",
	"INSERT INTO provisioning_params VALUES(default, 'softkey7', 'boolean', 'Soft Key 7 Enable', '', 'false', '307', '', null)",
	"INSERT INTO provisioning_params VALUES(default, 'softkey8', 'boolean', 'Soft Key 8 Enable', '', 'false', '308', '', null)",
	"INSERT INTO provisioning_params VALUES(default, 'softkey9', 'boolean', 'Soft Key 9 Enable', '', 'false', '309', '', null)",
	"INSERT INTO provisioning_params VALUES(default, 'softkey10', 'boolean', 'Soft Key 10 Enable', '', 'false', '310', '', null)",
	"INSERT INTO provisioning_params VALUES(default, 'databaseversion', 'text', 'Provisioning Database Version', 'Do not edit or delete!', '13', '999', '', null)",
}

-- All of the (non-default) parameter values for all devices are stored here
mymodule.provisioning_values = {
	-- device_id is a device id from provisioning_device_seq and param_id is a param id from provisioning_params
	"CREATE TABLE provisioning_values (device_id INTEGER, group_name VARCHAR(255), param_id INTEGER REFERENCES provisioning_params, value VARCHAR(255))",
	"CREATE INDEX values_device_idx ON provisioning_values (device_id)",
	"CREATE INDEX values_group_idx ON provisioning_values (group_name)",
	"CREATE INDEX values_param_idx ON provisioning_values (param_id)",
	"CREATE UNIQUE INDEX values_device_group_param_idx ON provisioning_values (device_id, group_name, param_id)",
	-- Need to enforce that group_name is a valid name in provisioning_groups (cannot use foreign key because name is not unique in provisioning_groups)
	-- Better yet, we'll check for a valid combination of device, group, and param (using triggers)
	-- Theoretically should also check on updates and updates of devices, groups, or parameters, but not going to bother
	"CREATE OR REPLACE FUNCTION check_valid_param() RETURNS TRIGGER AS $$ \
	        BEGIN \
			PERFORM * from devices_to_classes d join provisioning_classes using(class_id) join classes_to_param_groups using(class_id) join provisioning_groups g using(group_id) join param_groups_to_params p using(group_id) where d.device_id=NEW.device_id and g.name=NEW.group_name and p.param_id=NEW.param_id; \
			IF NOT FOUND THEN \
				RAISE EXCEPTION 'Invalid combination of device, group, and parameter'; \
			END IF; \
			RETURN NEW; \
		END; \
	$$ LANGUAGE plpgsql",
	"CREATE TRIGGER valid_param_trigger BEFORE INSERT ON provisioning_values FOR EACH ROW EXECUTE PROCEDURE check_valid_param()",
}

-- List of options for parameters
mymodule.provisioning_options = {
	"CREATE TABLE provisioning_options (param_id INTEGER REFERENCES provisioning_params, label VARCHAR(255), value VARCHAR(255), seq INTEGER)",
	"CREATE INDEX options_param_idx ON provisioning_options (param_id)",
	"CREATE UNIQUE INDEX options_param_label_idx ON provisioning_options (param_id, label)",
	-- Templates
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='template'), 'Polycom', '/etc/provisioning/templates/polycom-template.lua', '1')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='template'), 'Linksys ATA', '/etc/provisioning/templates/linksysata-template.lua', '2')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='template'), 'Snom', '/etc/provisioning/templates/snom-template.lua', '3')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='template'), 'Cisco ATA', '/etc/provisioning/templates/ciscoata-template.lua', '4')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='template'), 'Algo 8180', '/etc/provisioning/templates/algo8180-template.lua', '5')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='template'), 'CyberData', '/etc/provisioning/templates/cyberdata-template.lua', '6')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='template'), 'Grandstream', '/etc/provisioning/templates/grandstream-template.lua', '7')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='polycomringtone'), 'Silent Ring', '1', '1')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='polycomringtone'), 'Low Trill', '2', '2')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='polycomringtone'), 'Low Double Trill', '3', '3')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='polycomringtone'), 'Medium Trill', '4', '4')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='polycomringtone'), 'Medium Double Trill', '5', '5')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='polycomringtone'), 'High Trill', '6', '6')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='polycomringtone'), 'High Double Trill', '7', '7')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='polycomringtone'), 'Highest Trill', '8', '8')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='polycomringtone'), 'Highest Double Trill', '9', '9')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='polycomringtone'), 'Beeble', '10', '10')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='polycomringtone'), 'Triplet', '11', '11')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='polycomringtone'), 'Ringback-style', '12', '12')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='polycomringtone'), 'Low Trill Precedence', '13', '13')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='polycomringtone'), 'Ring Splash', '14', '14')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='provpollingtime'), '01:00', '01:00', '1')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='provpollingtime'), '02:00', '02:00', '2')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='provpollingtime'), '03:00', '03:00', '3')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='provpollingtime'), '04:00', '04:00', '4')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='provpollingtime'), '05:00', '05:00', '5')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='provpollingtime'), '06:00', '06:00', '6')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='provpollingtime'), '07:00', '07:00', '7')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='provpollingtime'), '08:00', '08:00', '8')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='provpollingtime'), '09:00', '09:00', '9')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='provpollingtime'), '10:00', '10:00', '10')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='provpollingtime'), '11:00', '11:00', '11')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='provpollingtime'), '12:00', '12:00', '12')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='provpollingtime'), '13:00', '13:00', '13')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='provpollingtime'), '14:00', '14:00', '14')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='provpollingtime'), '15:00', '15:00', '15')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='provpollingtime'), '16:00', '16:00', '16')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='provpollingtime'), '17:00', '17:00', '17')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='provpollingtime'), '18:00', '18:00', '18')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='provpollingtime'), '19:00', '19:00', '19')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='provpollingtime'), '20:00', '20:00', '20')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='provpollingtime'), '21:00', '21:00', '21')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='provpollingtime'), '22:00', '22:00', '22')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='provpollingtime'), '23:00', '23:00', '23')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='provpollingtime'), '00:00', '00:00', '24')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='displaylang'), 'English (United States)', 'English_United_States', '1')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='displaylang'), 'English (Canada)', 'English_Canada', '2')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='displaylang'), 'English (United Kingdom)', 'English_United_Kingdom', '3')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='displaylang'), 'Chinese (China)', 'Chinese_China', '4')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='displaylang'), 'Chinese (Taiwan)', 'Chinese_Taiwan', '5')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='displaylang'), 'Danish', 'Danish_Denmark', '6')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='displaylang'), 'Dutch', 'Dutch_Netherlands', '7')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='displaylang'), 'French (France)', 'French_France', '8')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='displaylang'), 'French (Canada)', 'French_Canada', '8')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='displaylang'), 'German', 'German_Germany', '9')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='displaylang'), 'Italian', 'Italian_Italy', '10')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='displaylang'), 'Japanese', 'Japanese_Japan', '11')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='displaylang'), 'Korean', 'Korean_Korea', '12')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='displaylang'), 'Norwegian', 'Norwegian_Norway', '13')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='displaylang'), 'Polish', 'Polish_Poland', '14')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='displaylang'), 'Portuguese', 'Portuguese_Portugal', '15')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='displaylang'), 'Russian', 'Russian_Russia', '16')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='displaylang'), 'Slovenian', 'Slovenian_Slovenia', '17')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='displaylang'), 'Spanish', 'Spanish_Spain', '18')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='displaylang'), 'Swedish', 'Swedish_Sweden', '19')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='displaylang'), 'Internal (US English)', '', '20')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='algoringtone'), 'Bell NA', 'bell-na.wav', '1')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='algoringtone'), 'Bell UK', 'bell-uk.wav', '2')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='algoringtone'), 'Buzzer', 'buzzer.wav', '3')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='algoringtone'), 'Chime', 'chime.wav', '4')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='algoringtone'), 'Dogs', 'dogs.wav', '5')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='algoringtone'), 'Gong', 'gong.wav', '6')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='algoringtone'), 'Page Notify', 'page-notif.wav', '7')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='algoringtone'), 'Warble Low', 'warble1-low.wav', '8')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='algoringtone'), 'Warble Med', 'warble2-med.wav', '9')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='algoringtone'), 'Warble High', 'warble3-high.wav', '10')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='algoringtone'), 'Warble Trill', 'warble4-trill.wav', '11')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='algopagetone'), 'None', 'None', '1')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='algopagetone'), 'Bell NA', 'bell-na.wav', '2')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='algopagetone'), 'Bell UK', 'bell-uk.wav', '3')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='algopagetone'), 'Buzzer', 'buzzer.wav', '4')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='algopagetone'), 'Chime', 'chime.wav', '5')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='algopagetone'), 'Dogs', 'dogs.wav', '6')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='algopagetone'), 'Gong', 'gong.wav', '7')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='algopagetone'), 'Page Notify', 'page-notif.wav', '8')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='algopagetone'), 'Warble Low', 'warble1-low.wav', '9')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='algopagetone'), 'Warble Med', 'warble2-med.wav', '10')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='algopagetone'), 'Warble High', 'warble3-high.wav', '11')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='algopagetone'), 'Warble Trill', 'warble4-trill.wav', '12')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='algoringvol'), '1', '1', '1')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='algoringvol'), '2', '2', '2')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='algoringvol'), '3', '3', '3')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='algoringvol'), '4', '4', '4')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='algoringvol'), '5', '5', '5')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='algoringvol'), '6', '6', '6')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='algoringvol'), '7', '7', '7')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='algoringvol'), '8', '8', '8')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='algoringvol'), '9', '9', '9')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='algoringvol'), '10', '10', '10')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='algopagevol'), '1', '1', '1')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='algopagevol'), '2', '2', '2')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='algopagevol'), '3', '3', '3')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='algopagevol'), '4', '4', '4')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='algopagevol'), '5', '5', '5')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='algopagevol'), '6', '6', '6')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='algopagevol'), '7', '7', '7')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='algopagevol'), '8', '8', '8')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='algopagevol'), '9', '9', '9')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='algopagevol'), '10', '10', '10')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='algospkmode'), 'Auto', '0', '0')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='algospkmode'), 'External', '1', '1')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='algospkmode'), 'Internal', '2', '2')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='algospkmode'), 'Disabled', '3', '3')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='cyberspeakervolume'), '0', '0', '0')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='cyberspeakervolume'), '1', '1', '1')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='cyberspeakervolume'), '2', '2', '2')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='cyberspeakervolume'), '3', '3', '3')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='cyberspeakervolume'), '4', '4', '4')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='cyberspeakervolume'), '5', '5', '5')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='cyberspeakervolume'), '6', '6', '6')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='cyberspeakervolume'), '7', '7', '7')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='cyberspeakervolume'), '8', '8', '8')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='cyberspeakervolume'), '9', '9', '9')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='cyberringvolume'), '0', '0', '0')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='cyberringvolume'), '1', '1', '1')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='cyberringvolume'), '2', '2', '2')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='cyberringvolume'), '3', '3', '3')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='cyberringvolume'), '4', '4', '4')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='cyberringvolume'), '5', '5', '5')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='cyberringvolume'), '6', '6', '6')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='cyberringvolume'), '7', '7', '7')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='cyberringvolume'), '8', '8', '8')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='cyberringvolume'), '9', '9', '9')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='cybermicgain'), '0', '0', '0')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='cybermicgain'), '1', '1', '1')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='cybermicgain'), '2', '2', '2')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='cybermicgain'), '3', '3', '3')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='cybermicgain'), '4', '4', '4')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='cybermicgain'), '5', '5', '5')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='cybermicgain'), '6', '6', '6')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='cybermicgain'), '7', '7', '7')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='cybermicgain'), '8', '8', '8')",
	"INSERT INTO provisioning_options VALUES((SELECT param_id FROM provisioning_params WHERE name='cybermicgain'), '9', '9', '9')",
}

-- List of getfile requests
mymodule.provisioning_requests = {
	"CREATE TABLE provisioning_requests (mac VARCHAR(12) unique, ip VARCHAR(15), agent VARCHAR(255), date timestamp)",
}

return mymodule
